# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

"""Utils for calibrating Diffraction Patterns."""

import numpy as np
import pint

from diffsims.utils.sim_utils import get_electron_wavelength
import hyperspy.api as hs
from hyperspy.axes import UniformDataAxis
from hyperspy.misc.utils import DictionaryTreeBrowser

from pyxem.utils.indexation_utils import index_dataset_with_template_rotation
from pyxem.utils._azimuthal_integrations import _get_control_points, _get_factors
from pyxem.utils._deprecated import deprecated
from typing import Union

Number = Union[int, float]

ureg = pint.UnitRegistry()

unit_equivalents = {
    "nm^-1": "nm$^{-1}$",
    "A^-1": r"$\AA^{-1}$",
    "$A^{-1}$": r"$\AA^{-1}$",
    "k_nm^-1": "nm$^{-1}$",
    "k_A^-1": r"$\AA^{-1}$",
    "nm-1": "nm$^{-1}$",
    "A-1": r"$\AA^{-1}$",
}


def value2standard_units(
    value: Union[str, pint.Quantity, Number],
    standard_unit: Union[str, pint.Unit] = None,
) -> pint.Quantity:
    """
    Convert a string representation of units to a standard unit.

    This takes a string like "40 mm" and will convert it to SI units 4E-3 m.
    If the string is not recognized, it will raise a ValueError.

    Parameters
    ----------
    value: str
        The string representation of the units, e.g. "40 mm", "1 cm", "100 um".
    standard_unit: str,
        The standard unit to convert to. If not provided, it will use the
        unit from the value if it is a pint.Quantity, or raise an error if
        the value is a number.

    Returns
    -------
    pint.Quantity
        The value converted to the standard unit.
    """
    if isinstance(value, pint.Quantity):
        return value.to(standard_unit) if standard_unit else value
    elif isinstance(value, (int, float)):
        if standard_unit is None:
            raise ValueError("If value is a number, standard_unit must be provided.")
        return ureg.Quantity(value, standard_unit)
    elif isinstance(value, str):
        quantity = ureg.Quantity(value)
        if standard_unit is not None:
            return quantity.to(standard_unit)
        return quantity


def value2node(
    value: Union[str, pint.Quantity, Number],
    metadata: DictionaryTreeBrowser,
    metadata_key: str,
    standard_unit: Union[str, pint.Unit] = None,
) -> pint.Quantity:
    """Convert a value to a pint.Quantity and store it in the metadata along with
    a unit.

    Parameters
    ----------
    value:
        The value to convert. This can be a string, pint.Quantity, or a number.
    metadata: hs.metadata.Metadata
        The metadata object to store the value in.
    metadata_key:
        The metadata key to store the value in.
    standard_unit:
        The standard unit to convert to. If not provided, it will use the unit from the value
        if it is a pint.Quantity, or raise an error if the value is a number.
    """

    value = value2standard_units(value, standard_unit)
    metadata.set_item(metadata_key, value.m)
    metadata_key_units = f"{metadata_key}_units"
    metadata.set_item(metadata_key_units, str(value.units))
    return value.m


class Calibration:
    """
    This is a class to hold the calibration parameters for some dataset
    and the methods for calibrating the dataset.

    It is designed to hold information about the affine transformation
    and the mask to apply to the data for something like a beam stop.

    There are 2 ways to set the calibration:
    1. You can set the calibration with a known reciprocal space pixel size.  This will assume
       a flat ewald sphere.

    2. You can set the calibration directly with a known real pixel size, beam energy/wavelength,
       and detector distance. This is the most accurate method but requires detailed calibration
       information that is not trivial to acquire.  In most cases option 1 is sufficient.

    If you set the pixel size with `hyperSpy.axes_manager.signal_axes[0].scale = 0.1`
    it will default to the first method. The underlying data will be stored in the metadata
    of the signal to be used later.
    """

    def __init__(
        self,
        signal,
    ):
        """A class to hold the parameters for an Azimuthal Integrator.

        Parameters
        ----------
        affine: (3x3)
            The affine transformation to apply to the data
        mask:
            A boolean array to be added to the integrator.
        """
        self.signal = signal
        self._mrad_scale = None

    def __call__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def _repr_html_(self):
        """Return an HTML representation of the Calibration object."""

        pixel_size = np.array(getattr(self, "physical_pixel_size", "N/A"))
        if isinstance(pixel_size, np.ndarray):
            pixel_size = f"{pixel_size * 1e6}"

        camera_length = getattr(self, "camera_length", 0)
        if camera_length is None:
            camera_length = 0
        camera_length = f"{camera_length:.2f}" if camera_length > 0 else "N/A"

        html = f"""
            <table style="border-collapse: collapse; width: 50%;">
            <caption style="font-weight: bold; font-size: 1.2em;">Calibration</caption>
            <tr><th style="border: 1px solid #ccc; padding: 4px;">Attribute</th>
                <th style="border: 1px solid #ccc; padding: 4px;">Value</th></tr>
            <tr><td style="border: 1px solid #ccc; padding: 4px;">Scale</td>
                <td style="border: 1px solid #ccc; padding: 4px;">{self.scale}</td></tr>
            <tr><td style="border: 1px solid #ccc; padding: 4px;">Units</td>
                <td style="border: 1px solid #ccc; padding: 4px;">{self.units}</td></tr>
            <tr><td style="border: 1px solid #ccc; padding: 4px;">Center</td>
                <td style="border: 1px solid #ccc; padding: 4px;">{self.center}</td></tr>
            <tr><td style="border: 1px solid #ccc; padding: 4px;">Beam Energy (kV)</td>
                <td style="border: 1px solid #ccc; padding: 4px;">{getattr(self, 'beam_energy', 'N/A')}</td></tr>
            <tr><td style="border: 1px solid #ccc; padding: 4px;">Pixel Size (um)</td>
                <td style="border: 1px solid #ccc; padding: 4px;">{pixel_size}</td></tr>
            <tr><td style="border: 1px solid #ccc; padding: 4px;">Camera Length (mm)</td>
                <td style="border: 1px solid #ccc; padding: 4px;">{camera_length}</td></tr>
        </table>
        """
        return html

    def convert_signal_units(self, new_unit):
        """Change the units of the signal axes.

        Parameters
        ----------
        unit: str
            The new unit to set for the signal axes. Options are:
            - 'px': pixels
            - 'mrad': milliradians
            - 'nm^-1': nanometers^-1
            - 'A^-1': Angstroms^-1
        """
        if new_unit in unit_equivalents.keys():
            new_unit = unit_equivalents[new_unit]

        u = self.units[0]
        if u in unit_equivalents.keys():
            u = unit_equivalents[u]
        if u in ["mrad", "nm$^{-1}$", r"$\AA^{-1}$"]:
            # If the units are already in mrad, nm^-1, or A^-1, we need to set the mrad scale
            self.set_mrad_scale_from_calibration()

        if new_unit == "px":
            # If the new unit is pixels, we need to set the scale to 1
            scale = 1
            offset = np.array(self.center) * -1
        elif new_unit == "mrad":
            scale = np.array(self._mrad_scale)
            offset = np.array(self.center) * -scale
        elif new_unit == "nm$^{-1}$":
            scale = np.tan(self._mrad_scale / 1000) * (1 / self.wavelength) * 10
            offset = np.array(self.center) * -scale
        elif new_unit == r"$\AA^{-1}$":
            scale = np.tan(self._mrad_scale / 1000) * (1 / self.wavelength)
            offset = np.array(self.center) * -scale
        else:
            raise ValueError(
                f"Cannot change units to {new_unit}. "
                "Must be one of 'px', 'mrad', 'nm$^{-1}$', or r'$\\AA^{-1}$'."
            )
        # Set the new units and scale
        self.signal.axes_manager.signal_axes.set(
            units=new_unit, scale=scale, offset=offset
        )

    def set_mrad_scale_from_calibration(self):
        """
        Set the mrad scale from the calibration parameters.
        """
        scales = []
        for u, s in zip(self.units, self.scale):
            if u in unit_equivalents.keys():
                u = unit_equivalents[u]
            if u == "mrad":
                scales.append(s / 1000)  # convert to rad
            elif u == "nm$^{-1}$":
                angle_one_pix = np.arctan2(s / 10, 1 / self.wavelength)
                scales.append(np.tan(angle_one_pix))
            elif u == r"$\AA^{-1}$":
                angle_one_pix = np.arctan2(s, 1 / self.wavelength)
                scales.append(np.tan(angle_one_pix))
            else:
                scales = self._mrad_scale / 1000
                if scales is None:
                    raise ValueError(
                        f"Cannot set mrad scale from calibration unless already set if the units are {u}. "
                        "Must be one of 'mrad', 'nm$^{-1}$', or r'$\\AA^{-1}$'."
                    )
        self._mrad_scale = np.array(scales) * 1000

    def set_camera_length_from_calibration(self):
        """Sets the camera length from the calibration parameters.

        Many times the camera length is not exactly known.  Values from the
        microscope are often not exactly accurate.  This method will set the
        camera length based on the units and the scale currently set for
        some signal. This does require that the detector pixel size and the
        beam energy are set as well.
        """
        if not self.flat_ewald:
            raise ValueError(
                "Cannot set camera length from calibration if the ewald sphere is not flat."
            )
        units = self.units[0]
        if units in unit_equivalents.keys():
            units = unit_equivalents[units]
        if units == "mrad":
            camera_length = self.physical_pixel_size / np.tan(self.scale[0] / 1000)
            self.camera_length = camera_length
        elif units == "nm$^{-1}$":
            angle_one_pix = np.arctan2(self.scale[0] / 10, 1 / self.wavelength)
            camera_length = self.physical_pixel_size / np.tan(angle_one_pix)
            self.camera_length = camera_length
        elif units == r"$\AA^{-1}$":
            angle_one_pix = np.arctan2(self.scale[0], 1 / self.wavelength)
            camera_length = self.physical_pixel_size / np.tan(angle_one_pix)
            self.camera_length = camera_length
        elif units == "px":
            pass
        else:
            raise ValueError(
                f"Cannot set camera length from calibration if the units are {self.units[0]}"
            )

    @property
    def affine(self):
        """Set the affine transformation to apply to the data."""
        return self.signal.metadata.get_item("General.affine_transformation")

    @affine.setter
    def affine(self, affine):
        self.signal.metadata.set_item("General.affine_transformation", affine)

    @property
    def mask(self):
        """Set the mask to apply to the data."""
        return self.signal.metadata.get_item("General.mask")

    @mask.setter
    def mask(self, mask):
        self.signal.metadata.set_item("General.mask", mask)

    @property
    def scale(self):
        """Return the scale in units of the axes."""
        if self.flat_ewald:
            return [s.scale for s in self.signal.axes_manager.signal_axes]
        else:
            return None

    @scale.setter
    def scale(self, scale):
        """Set the scale in units of the axes."""
        if not self.flat_ewald:
            raise ValueError("Scale can only be set if the ewald sphere is flat")
        center = self.center  # save the center
        for ax, c in zip(self.signal.axes_manager.signal_axes, center):
            ax.scale = scale
            ax.offset = -c * scale

    @property
    def units(self):
        return [s.units for s in self.signal.axes_manager.signal_axes]

    @units.setter
    def units(self, units):
        # maybe consider converting the scales as well?
        for ax in self.signal.axes_manager.signal_axes:
            ax.units = units

    @property
    def camera_length(self):
        """
        The camera length in meters.
        This is the "distance" from the sample to the detector.
        """
        return self.signal.metadata.get_item("Acquisition_instrument.TEM.camera_length")

    @camera_length.setter
    def camera_length(self, camera_length):
        """Set the camera length in meters."""
        if camera_length is None:
            self.signal.metadata.remove_item("Acquisition_instrument.TEM.camera_length")
            return
        value2node(
            camera_length,
            self.signal.metadata,
            "Acquisition_instrument.TEM.camera_length",
            standard_unit="m",
        )

    @property
    def physical_pixel_size(self):
        """
        The physical pixel size of the detector in m
        """
        return self.signal.metadata.get_item(
            "Acquisition_instrument.TEM.physical_pixel_size"
        )

    @physical_pixel_size.setter
    def physical_pixel_size(self, pixel_size):
        """Set the physical pixel size of the detector in m."""
        if pixel_size is None:
            self.signal.metadata.remove_item(
                "Acquisition_instrument.TEM.physical_pixel_size"
            )
            return
        value2node(
            pixel_size,
            self.signal.metadata,
            "Acquisition_instrument.TEM.physical_pixel_size",
            standard_unit="m",
        )

    @property
    def convergence_angle(self):
        """
        The convergence angle of the beam in mrad.
        """
        return self.signal.metadata.get_item(
            "Acquisition_instrument.TEM.convergence_angle"
        )

    @convergence_angle.setter
    def convergence_angle(self, convergence_angle):
        """Set the convergence angle of the beam in mrad."""
        if convergence_angle is None:
            self.signal.metadata.remove_item(
                "Acquisition_instrument.TEM.convergence_angle"
            )
            return
        value2node(
            convergence_angle,
            self.signal.metadata,
            "Acquisition_instrument.TEM.convergence_angle",
            standard_unit="mrad",
        )

    @property
    def wavelength(self):
        """The electron wavelength in A^-1."""
        wavelength = self.signal.metadata.get_item(
            "Acquisition_instrument.TEM.wavelength"
        )
        if isinstance(wavelength, pint.Quantity):
            return wavelength.m
        else:
            return wavelength

    @wavelength.setter
    def wavelength(self, wavelength):
        """Set the electron wavelength in A^-1."""
        if wavelength is None:
            self.signal.metadata.remove_item("Acquisition_instrument.TEM.wavelength")
            return
        wavelength = value2standard_units(wavelength, standard_unit="angstrom^-1")
        self.signal.metadata.set_item(
            "Acquisition_instrument.TEM.wavelength", wavelength
        )

    @property
    def beam_energy(self):
        """The beam energy in kV."""
        return self.signal.metadata.get_item("Acquisition_instrument.TEM.beam_energy")

    @beam_energy.setter
    def beam_energy(self, beam_energy):
        """Set the beam energy in kV."""
        energy = value2node(
            beam_energy,
            self.signal.metadata,
            "Acquisition_instrument.TEM.beam_energy",
            standard_unit="kV",
        )

        self.signal.metadata.set_item(
            "Acquisition_instrument.TEM.wavelength",
            get_electron_wavelength(energy),
        )

    @property
    def detector_gain(self):
        return self.signal.metadata.get_item("Acquisition_instrument.TEM.detector_gain")

    @detector_gain.setter
    def detector_gain(self, gain):
        """
        Calibrate the signal with a known detector gain. This works by dividing the signal the average
        number of counts per electron.  It also sets the appropriate metadata for setting the color
        bar when plotting the signal.
        Parameters
        ----------
        gain: float
            The gain of the detector
        """
        if gain != 1:  # ignore for counted detectors
            self.signal.data = self.signal.data / gain
        self.signal.metadata.set_item(
            "Signal.quantity", "e$^-$"
        )  # set the quantity to be electrons
        self.signal.metadata.set_item("Acquisition_instrument.TEM.detector_gain", gain)

    def detector(
        self,
        pixel_size,
        detector_distance,
        beam_energy=None,
        wavelength=None,
        center=None,
        units="k_nm^-1",
    ):
        """
        Calibrate the signal with a known pixel size, detector distance, and beam energy/wavelength.


        This sets the signal axes to be a `~hyperspy.axes.DataAxis` instead of a UniformAxis to
        account for the Ewald sphere curvature. This is the most accurate method of calibration
        but requires a known beam energy/wavelength and detector distance or to calibration the
        experimental configuration with a known standard.

        Parameters
        ----------
        pixel_size: float
            The pixel size in meter.
        detector_distance: float
            The detector distance in meter.
        beam_energy: float
            The beam energy in kV.
        wavelength: float
            The beam wavelength in A^-1.
        center: list
            The center of the signal in pixels.
        unit: str
            The unit to calculate the radial extent with.
        """

        if beam_energy is None and wavelength is None:
            wavelength = self.wavelength
            if wavelength is None:
                raise ValueError("Must provide either beam_energy or wavelength")
        elif beam_energy is not None:
            wavelength = get_electron_wavelength(beam_energy)
            self.beam_energy = beam_energy
        self.wavelength = wavelength
        # We need these values to properly set the axes
        self.signal.metadata.set_item(
            "Acquisition_instrument.TEM.physical_pixel_size", pixel_size
        )
        self.signal.metadata.set_item(
            "Acquisition_instrument.TEM.detector_distance", detector_distance
        )

        unit_factors = {
            "k_nm^-1": 10,
            "k_A^-1": 1,
            "q_nm^-1": 10 / (2 * np.pi),
            "q_A^-1": 1 / (2 * np.pi),
        }
        if units not in unit_factors.keys():
            raise ValueError(
                f"Unit {units} not recognized. Must be one of {unit_factors.keys()}"
            )

        if center is None:
            center = [(ax.size - 1) / 2 for ax in self.axes]

        def translate_pixel_coords(px: np.ndarray) -> np.ndarray:
            coord = pixel_size * px
            angle = np.arctan(coord / detector_distance)
            return np.sin(angle) * (1 / wavelength) * unit_factors[units]

        x_pixels = np.arange(self.shape[0]) - center[0]
        y_pixels = np.arange(self.shape[1]) - center[1]

        x_axes = translate_pixel_coords(x_pixels)
        y_axes = translate_pixel_coords(y_pixels)

        for ax, axis in zip(self.signal.axes_manager.signal_axes, [x_axes, y_axes]):
            if isinstance(ax, UniformDataAxis):
                ax.convert_to_non_uniform_axis()
            ax.axis = axis
        self.units = units

    @property
    def shape(self):
        return self.signal.axes_manager.signal_shape[::-1]

    @property
    def flat_ewald(self):
        """If the ewald sphere is flat return True"""
        return isinstance(self.signal.axes_manager.signal_axes[0], UniformDataAxis)

    def __repr__(self):
        rep_str = f"Calibration for {self.signal}, Ewald sphere: "
        if self.flat_ewald:
            rep_str += "flat"
        else:
            rep_str += f"curved"
        rep_str += f", shape: {self.shape}, affine: {self.affine is not None},"
        rep_str += f" mask: {self.mask is not None}"
        return rep_str

    @property
    def axes(self):
        return [ax.axis for ax in self.signal.axes_manager.signal_axes][::-1]

    @property
    def pixel_extent(self):
        """Return an array with axes [x/y, left/right, pixel_extent], as follows:
        [
            # x axis
            [
                # left
                [boundary 1, boundary 2 ...],
                # right
                [boundary 1, boundary 2 ...],
            ],
            # y axis
            [
                # left
                [boundary 1, boundary 2 ...],
                # right
                [boundary 1, boundary 2 ...],
            ],
        ]
        """
        if self.flat_ewald:
            left_scales = self.scale
            right_scales = self.scale
        else:
            x_scales = self.axes[0][1:] - self.axes[0][:-1]
            x_scales = np.pad(x_scales, 1, mode="edge")
            y_scales = self.axes[1][1:] - self.axes[1][:-1]
            y_scales = np.pad(y_scales, 1, mode="edge")
            left_scales = [
                x_scales[:-1],
                y_scales[:-1],
            ]
            right_scales = [
                x_scales[1:],
                y_scales[1:],
            ]

        extents = []
        for ax, left_scale, right_scale in zip(self.axes, left_scales, right_scales):
            left = ax - left_scale / 2
            right = ax + right_scale / 2
            extent = np.stack((left, right))
            extents.append(extent)
        return extents

    def get_slices2d(self, npt, npt_azim, radial_range=None, azimuthal_range=None):
        """Get the slices and factors for some image that can be used to
        slice the image for 2d integration.

        Parameters
        ----------
        npt: int
            The number of radial points
        npt_azim:
            The number of azimuthal points
        """
        radial_range = self._get_radial_range(radial_range)
        if azimuthal_range is None:
            azimuthal_range = (-np.pi, np.pi)
        # Get the slices and factors for the integration
        slices, factors, factors_slice = self._get_slices_and_factors(
            npt, npt_azim, radial_range, azimuthal_range
        )
        return slices, factors, factors_slice, radial_range

    def _get_radial_range(self, radial_range=None):
        if radial_range is None:
            from itertools import combinations

            edges = np.reshape(
                [
                    [ax_ext[0][0] ** 2, ax_ext[1][-1] ** 2]
                    for ax_ext in self.pixel_extent
                ],
                -1,
            )
            max_range = np.max(
                np.power(np.sum(list(combinations(edges, 2)), axis=1), 0.5)
            )
            radial_range = (0, max_range)
        return radial_range

    def get_slices1d(self, npt, radial_range=None):
        """Get the slices and factors for some image that can be used to
        slice the image for 1d integration.

        Parameters
        ----------
        npt: int
            The number of radial points
        radial_range: tuple
            The range of the radial extent

        Returns
        -------
        indexes: np.ndarray (n, 2)
            The indexes of the pixels to integrate flattened
        factors: np.ndarray (n)
            The factors(representing the pixel fraction) to multiply each pixel value by
        factor_slices: np.ndarray (npt+1)
            The start and end index of the factors for each slice such that for some
            slice i, factors[factor_slices[i]:factor_slices[i+1]] is the factors for that radial slice
        radial_range: tuple
            The range of the radial extent used for the integration

        """
        # reuse the 2d method as it is actually fairly fast
        # and can be much faster if we spend time with numba
        npt_azim = 360  # approximate a circle with a 360-gon... Using a circle is harder/ not much better

        slices, factors, factors_slice, radial_range = self.get_slices2d(
            npt, npt_azim, radial_range
        )

        # convert into 1d slices
        indexes = []
        facts = []
        for i in range(npt):
            test = np.zeros(self.shape)
            for j in range(npt_azim):
                ind = i * npt_azim + j
                sl = slices[ind]
                test[sl[0] : sl[2], sl[1] : sl[3]] += factors[
                    factors_slice[ind][0] : factors_slice[ind][1]
                ].reshape((sl[2] - sl[0], sl[3] - sl[1]))
            inds = np.argwhere(test)
            indexes.append(inds)
            facts.append(test[inds[:, 0], inds[:, 1]])
        factor_slices = np.cumsum(
            [
                0,
            ]
            + [len(f) for f in facts]
        )
        indexes, facts = np.vstack(indexes), np.hstack(facts)
        return indexes, facts, factor_slices, radial_range

    def _get_slices_and_factors(self, npt, npt_azim, radial_range, azimuthal_range):
        # In `_get_control_points`, positive x-direction is downwards, and positive y is to the right.
        # This is 90 degrees off from the pyxem definition:
        # https://pyxem.readthedocs.io/en/stable/tutorials/pyxem-demos/13%20Conventions.html
        # As the azimuthal integration performed with a `Calibration`-object should align with
        # the $X_L$ / $Y_L$-definitions in pyxem, we add pi/2 to the azimuthal range.
        # This aligns the azimuthal angle 0 to $X_L$.
        azimuthal_range = (
            azimuthal_range[0] + np.pi / 2,
            azimuthal_range[1] + np.pi / 2,
        )

        # get the points which bound each azimuthal pixel
        control_points = _get_control_points(
            npt, npt_azim, radial_range, azimuthal_range, self.affine
        )

        # get the min and max indices for each control point using the
        pixel_ext_x, pixel_ext_y = self.pixel_extent
        min_x = (
            np.min(
                np.searchsorted(
                    pixel_ext_x[1, :], control_points[:, :, 0], side="left"
                ),
                axis=1,
            ).astype(int)
            - 1
        )
        min_y = (
            np.min(
                np.searchsorted(
                    pixel_ext_y[1, :], control_points[:, :, 1], side="left"
                ),
                axis=1,
            ).astype(int)
            - 1
        )

        max_x = np.max(
            np.searchsorted(pixel_ext_x[0, :], control_points[:, :, 0], side="right"),
            axis=1,
        ).astype(int)
        max_y = np.max(
            np.searchsorted(pixel_ext_y[0, :], control_points[:, :, 1], side="right"),
            axis=1,
        ).astype(int)
        # Note that if a point is outside the range of the axes it will be set to the
        # maximum value of the axis+1 and 0 if it is below the minimum value of the axis.
        slices = np.array(
            [[mx, my, mxx, myy] for mx, my, mxx, myy in zip(min_x, min_y, max_x, max_y)]
        )
        max_y_ind = len(self.axes[1])
        max_x_ind = len(self.axes[0])

        # set the slices to be within the range of the axes.  If the entire slice is outside
        # the range of the axes then set the slice to be the maximum value of the axis
        slices[slices[:, 0] > max_x_ind, 0] = max_x_ind
        slices[slices[:, 1] > max_y_ind, 1] = max_y_ind
        slices[slices[:, 2] > max_x_ind, 2] = max_x_ind
        slices[slices[:, 3] > max_y_ind, 3] = max_y_ind

        slices[slices[:, 0] < 0, 0] = 0
        slices[slices[:, 1] < 0, 1] = 0
        slices[slices[:, 2] < 0, 2] = 0
        slices[slices[:, 3] < 0, 3] = 0

        factors, factors_slice = _get_factors(control_points, slices, self.pixel_extent)
        return slices, factors, factors_slice

    @property
    def center(self):
        """Return the center in pixels"""
        if self.flat_ewald:
            return [-s.offset / s.scale for s in self.signal.axes_manager.signal_axes]
        else:
            return [
                np.argmin(np.abs(s.axis)) for s in self.signal.axes_manager.signal_axes
            ]

    @center.setter
    def center(self, center=None):
        """Set the center in pixels"""
        if not self.flat_ewald:
            raise ValueError(
                "To set the center of a curved ewald sphere "
                "use the s.calibration.detector method"
            )
        if center is None:
            for ax in self.signal.axes_manager.signal_axes:
                ax.offset = -ax.scale * ((ax.size - 1) / 2)
        else:
            for ax, off in zip(self.signal.axes_manager.signal_axes, center):
                ax.offset = -off * ax.scale


@deprecated(
    since="0.18.0",
    removal="1.0.0",
    alternative="pyxem.signals.Diffraction2D.calibration",
)
def find_diffraction_calibration(
    patterns,
    calibration_guess,
    library_phases,
    lib_gen,
    size,
    max_excitation_error=0.01,
    **kwargs,
):
    """Finds the diffraction calibration for a pattern or set of patterns by maximizing correlation scores.

    Parameters
    ----------
    patterns : hyperspy.signals.Signal2D
        Diffration patterns to be iteratively matched to find maximum correlation scores.
    calibration_guess : float
        Inital value for the diffraction calibration in inverse Angstoms per pixel
    library_phases : diffsims.libraries.StructureLibrary
        Dictionary of structures and associated orientations for which
        electron diffraction is to be simulated.
    lib_gen : diffsims.generators.DiffractionLibraryGenerator
        Computes a library of electron diffraction patterns for specified atomic
        structures and orientations.  Used to create the DiffractionLibrary.
    size : integer
        How many different steps to test for the first two iterations. These steps have a size of 1% of the calibration guess.
    max_excitation_error : float
        Maximum exacitation error.  Default is 0.01.
    kwargs
        Keyword arguments passed to :meth:`index_dataset_with_template_rotation`.

    Returns
    -------
    mean_cal : float
        Mean of calibrations found for each pattern.
    full_corrlines : numpy.ndarray
        Gives the explicit correlation vs calibration values. Shape:(size*2 + 20, 2 , number of patterns)
    found_cals : numpy.ndarray
        List of optimal calibration values for each pattern. Shape:(number of patterns)
    """

    images = patterns

    num_patterns = images.data.shape[0]
    found_cals = np.full((num_patterns,), calibration_guess)
    full_corrlines = np.zeros((0, 2, num_patterns))

    stepsize = 0.01 * calibration_guess
    # first set of checks
    corrlines = _calibration_iteration(
        images,
        calibration_guess,
        library_phases,
        lib_gen,
        stepsize,
        size,
        num_patterns,
        max_excitation_error,
        **kwargs,
    )
    full_corrlines = np.append(full_corrlines, corrlines, axis=0)

    # refined calibration checks
    calibration_guess = full_corrlines[
        full_corrlines[:, 1, :].argmax(axis=0), 0, 0
    ].mean()
    corrlines = _calibration_iteration(
        images,
        calibration_guess,
        library_phases,
        lib_gen,
        stepsize,
        size,
        num_patterns,
        max_excitation_error,
        **kwargs,
    )
    full_corrlines = np.append(full_corrlines, corrlines, axis=0)

    # more refined calibration checks with smaller step
    stepsize = 0.001 * calibration_guess
    size = 20
    calibration_guess = full_corrlines[
        full_corrlines[:, 1, :].argmax(axis=0), 0, 0
    ].mean()

    corrlines = _calibration_iteration(
        images,
        calibration_guess,
        library_phases,
        lib_gen,
        stepsize,
        size,
        num_patterns,
        max_excitation_error,
        **kwargs,
    )
    full_corrlines = np.append(full_corrlines, corrlines, axis=0)
    found_cals = full_corrlines[full_corrlines[:, 1, :].argmax(axis=0), 0, 0]

    mean_cal = found_cals.mean()
    return mean_cal, full_corrlines, found_cals


def _calibration_iteration(
    images,
    calibration_guess,
    library_phases,
    lib_gen,
    stepsize,
    size,
    num_patterns,
    max_excitation_error,
    **kwargs,
):
    """For use in find_diffraction_calibration.  Controls the iteration of _create_check_diflib over a set of steps.

    Parameters
    ----------
    images : hyperspy.signals.Signal2D
        Diffration patterns to be iteratively matched to find maximum correlation scores.
    calibration_guess : float
        Inital value for the diffraction calibration in inverse Angstoms per pixel
    library_phases : diffsims.libraries.StructureLibrary
        Dictionary of structures and associated orientations for which
        electron diffraction is to be simulated.
    lib_gen : diffsims.generators.DiffractionLibraryGenerator
        Computes a library of electron diffraction patterns for specified atomic
        structures and orientations.  Used to create the DiffractionLibrary.
    stepsize : float
        Stepsize of iteration.
    size : integer
        How many different steps to test.
    num_patterns : integer
        Number of patterns.
    max_excitation_error : float
        Maximum exacitation error.  Default is 0.01.
    kwargs
        Keyword arguments passed to :meth:`index_dataset_with_template_rotation`.

    Returns
    -------
    corrlines : numpy.ndarray
    """
    corrlines = np.zeros((0, 2, num_patterns))
    temp_line = np.zeros((1, 2, num_patterns))
    cal_guess_greater = calibration_guess
    cal_guess_lower = calibration_guess
    for i in range(size // 2):
        temp_line[0, 0, :] = cal_guess_lower
        temp_line[0, 1, :] = _create_check_diflib(
            images,
            cal_guess_lower,
            library_phases,
            lib_gen,
            max_excitation_error,
            **kwargs,
        )
        corrlines = np.append(corrlines, temp_line, axis=0)

        temp_line[0, 0, :] = cal_guess_greater
        temp_line[0, 1, :] = _create_check_diflib(
            images,
            cal_guess_greater,
            library_phases,
            lib_gen,
            max_excitation_error,
            **kwargs,
        )
        corrlines = np.append(corrlines, temp_line, axis=0)

        cal_guess_lower = cal_guess_lower - stepsize
        cal_guess_greater = cal_guess_greater + stepsize

    return corrlines


def _create_check_diflib(
    images, calibration_guess, library_phases, lib_gen, max_excitation_error, **kwargs
):
    """For use in find_diffraction_calibration via _calibration_iteration.  Creates a new DiffractionLibrary from the inputs and then matches it the images.

    Parameters
    ----------
    images : hyperspy.signals.Signal2D
        Diffration patterns to be iteratively matched to find maximum correlation scores.
    calibration_guess : float
        Inital value for the diffraction calibration in inverse Angstoms per pixel
    library_phases : diffsims.libraries.StructureLibrary
        Dictionary of structures and associated orientations for which
        electron diffraction is to be simulated.
    lib_gen : diffsims.generators.DiffractionLibraryGenerator
        Computes a library of electron diffraction patterns for specified atomic
        structures and orientations.  Used to create the DiffractionLibrary.
    max_excitation_error : float
        Maximum exacitation error.  Default is 0.01.
    kwargs
        Keyword arguments passed to :meth:`pyxem.utisl.indexation_utils.index_dataset_with_template_rotation`.

    Returns
    -------
    correlations : numpy.ndarray
    """

    half_shape = (images.data.shape[-2] // 2, images.data.shape[-1] // 2)
    reciprocal_r = np.sqrt(half_shape[0] ** 2 + half_shape[1] ** 2) * calibration_guess
    diff_lib = lib_gen.get_diffraction_library(
        library_phases,
        calibration=calibration_guess,
        reciprocal_radius=reciprocal_r,
        half_shape=half_shape,
        with_direct_beam=False,
        max_excitation_error=max_excitation_error,
    )

    result, phasedict = index_dataset_with_template_rotation(images, diff_lib, **kwargs)
    correlations = result["correlation"][:, :, 0].flatten()
    return correlations



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
from pytest import approx
import numpy as np

from pyxem.signals import Diffraction2D
import pyxem.utils.radial_utils as ra
from pyxem.data.dummy_data import make_diffraction_test_data as mdtd


class TestRadialModule:
    def test_get_coordinate_of_min(self):
        array = np.ones((10, 10)) * 10
        x, y = 7, 5
        array[y, x] = 1  # In NumPy the order is [y, x]
        s = Diffraction2D(array)
        s.axes_manager[0].offset = 55
        s.axes_manager[1].offset = 50
        s.axes_manager[0].scale = 0.5
        s.axes_manager[1].scale = 0.4
        min_pos = ra.get_coordinate_of_min(s)
        # min_pos[0] (x) should be at (7*0.5) + 55 = 58.5
        # min_pos[1] (y) should be at (5*0.4) + 50 = 52.
        assert min_pos == (58.5, 52.0)


class TestGetAngleImageComparison:
    def setup_method(self):
        self.r0, self.r1 = 10, 20
        test_data0 = mdtd.MakeTestData(100, 100)
        test_data0.add_ring(50, 50, self.r0)
        test_data1 = mdtd.MakeTestData(100, 100)
        test_data1.add_ring(50, 50, self.r1)
        s0, s1 = test_data0.signal, test_data1.signal
        s0.axes_manager[0].offset, s0.axes_manager[1].offset = -50, -50
        s1.axes_manager[0].offset, s1.axes_manager[1].offset = -50, -50
        self.s0, self.s1 = s0, s1

    def test_different_angleN(self):
        s0, s1 = self.s0, self.s1
        for i in range(1, 10):
            ra.get_angle_image_comparison(s0, s1, angleN=i)

    def test_correct_radius(self):
        s0, s1, r0, r1 = self.s0, self.s1, self.r0, self.r1
        s = ra.get_angle_image_comparison(s0, s1, angleN=2)
        assert s.axes_manager.signal_shape == (100, 100)
        # Check that radius is correct by getting a line profile
        s_top = s.isig[0.0, :0.0]
        s_bot = s.isig[0.0, 0.0:]
        argmax0 = s_top.data.argmax()
        argmax1 = s_bot.data.argmax()
        assert abs(s_top.axes_manager[0].index2value(argmax0)) == r0
        assert abs(s_bot.axes_manager[0].index2value(argmax1)) == r1

    def test_different_signal_size(self):
        s0 = mdtd.MakeTestData(100, 100).signal
        s1 = mdtd.MakeTestData(100, 150).signal
        with pytest.raises(ValueError):
            ra.get_angle_image_comparison(s0, s1)

    def test_mask(self):
        s0, s1 = self.s0, self.s1
        s_no_mask = ra.get_angle_image_comparison(s0, s1)
        s_mask = ra.get_angle_image_comparison(s0, s1, mask_radius=40)
        assert s_no_mask.data.sum() != 0.0
        assert s_mask.data.sum() == 0.0


class TestFitEllipse:
    def setup_method(self):
        axis1, axis2 = 40, 70
        s = Diffraction2D(np.zeros((200, 220)))
        s.axes_manager[0].offset, s.axes_manager[1].offset = -100, -110
        xx, yy = np.meshgrid(s.axes_manager[0].axis, s.axes_manager[1].axis)
        ellipse_ring = mdtd._get_elliptical_ring(
            xx, yy, 0, 0, axis1, axis2, 0.8, lw_r=1
        )
        s.data += ellipse_ring
        self.s = s
        self.axis1, self.axis2 = axis1, axis2

    def test_private_functionality_uncovered_above(self):
        from pyxem.utils.radial_utils import _get_ellipse_parameters

        g_a_greater_than_c_b_zero = np.asarray([2, 0, 1, 0, 0, 0])
        g_a_less_than_c_b_zero = np.asarray([1, 0, 2, 0, 0, 0])
        for g in [g_a_greater_than_c_b_zero, g_a_less_than_c_b_zero]:
            _ = _get_ellipse_parameters(g)


class TestHolzCalibration:
    def test_get_holz_angle(self):
        wavelength = 2.51 / 1000
        lattice_parameter = 0.3905 * 2**0.5
        angle = ra._get_holz_angle(wavelength, lattice_parameter)
        assert approx(95.37805 / 1000) == angle

    def test_scattering_angle_to_lattice_parameter(self):
        wavelength = 2.51 / 1000
        angle = 95.37805 / 1000
        lattice_size = ra._scattering_angle_to_lattice_parameter(wavelength, angle)
        assert approx(0.55225047) == lattice_size



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

from hyperspy.signals import Signal2D
import numpy as np
from orix.quaternion import Rotation
import pytest

from pyxem.utils.indexation_utils import match_vectors
from pyxem.utils.indexation_utils import (
    index_dataset_with_template_rotation,
    results_dict_to_crystal_map,
)


def test_match_vectors(vector_match_peaks, vector_library):
    # Wrap to test handling of ragged arrays
    peaks = np.empty(1, dtype="object")
    peaks[0] = vector_match_peaks
    matches, rhkls = match_vectors(
        peaks,
        vector_library,
        mag_tol=0.1,
        angle_tol=0.1,
        index_error_tol=0.3,
        n_peaks_to_index=2,
        n_best=1,
    )
    assert len(matches) == 1
    np.testing.assert_allclose(matches[0].match_rate, 1.0)
    np.testing.assert_allclose(matches[0].rotation_matrix, np.identity(3), atol=0.1)
    np.testing.assert_allclose(matches[0].total_error, 0.03, atol=0.01)

    np.testing.assert_allclose(rhkls[0][0], [1, 0, 0])
    np.testing.assert_allclose(rhkls[0][1], [0, 2, 0])
    np.testing.assert_allclose(rhkls[0][2], [1, 2, 3])


def test_match_vector_total_error_default(vector_match_peaks, vector_library):
    matches, rhkls = match_vectors(
        vector_match_peaks,
        vector_library,
        mag_tol=0.1,
        angle_tol=0.1,
        index_error_tol=0.0,
        n_peaks_to_index=2,
        n_best=5,
    )
    assert len(matches) == 5
    np.testing.assert_allclose(matches[0][2], 0.0)  # match rate
    np.testing.assert_allclose(matches[0][1], np.identity(3), atol=0.1)
    np.testing.assert_allclose(matches[0][4], 1.0)  # error mean

    assert len(rhkls) == 0


@pytest.mark.filterwarnings("ignore:Property 'correlation' was expected")
def test_results_dict_to_crystal_map(test_library_phases_multi, test_lib_gen):
    """Test getting a :class:`orix.crystal_map.CrystalMap` from returns
    from :func:`index_dataset_with_template_rotation`.
    """
    # Map and signal shapes
    nav_shape = (2, 3)
    sig_shape = (80, 80)
    test_set = np.zeros(nav_shape + sig_shape)

    # Diffraction conditions
    diff_lib = test_lib_gen.get_diffraction_library(
        test_library_phases_multi,
        calibration=0.015,
        reciprocal_radius=1.18,
        half_shape=tuple(np.array(sig_shape) // 2),
        with_direct_beam=False,
    )

    # Simulate multi-phase results
    phase_id = np.zeros(nav_shape, dtype=int)
    phase_id[:, [0, 2]] = 1
    phase_names = list(diff_lib.keys())

    # Simulate patterns
    # TODO: Remove version check after diffsims 0.5.0 is released
    from packaging.version import Version
    import diffsims
    from scipy.ndimage import rotate

    diffsims_version = Version(diffsims.__version__)
    if diffsims_version > Version("0.4.2"):
        sim_kwargs = dict(shape=sig_shape, sigma=4)
    else:  # pragma: no cover
        sim_kwargs = dict(size=sig_shape[0], sigma=4)
    for idx in np.ndindex(*nav_shape):
        i = phase_id[idx]
        j = int(idx[1] / 2)
        test_pattern = diff_lib[phase_names[i]]["simulations"][
            j
        ].get_diffraction_pattern(**sim_kwargs)
        test_pattern = np.flipud(test_pattern)
        test_set[idx] = test_pattern

    test_set[0, 0] = rotate(test_set[0, 0], 30, reshape=False)
    # Perform template matching
    n_best = 3
    results, phase_dict = index_dataset_with_template_rotation(
        Signal2D(test_set),
        diff_lib,
        phases=phase_names,
        n_best=n_best,
    )

    # Extract various results once
    phase_id = results["phase_index"].reshape((-1, n_best))
    ori = np.deg2rad(results["orientation"].reshape((-1, n_best, 3)))
    orientation_result = results["orientation"][0, 0, 0, 0]

    assert (
        np.remainder(abs(orientation_result), 15) < 1
        or np.remainder(abs(orientation_result), 30) < 1
    )

    # Property names in `results`
    prop_names = ["correlation", "mirrored_template", "template_index"]

    # Only get the bast match when multiple phases match best to some
    # patterns
    xmap = results_dict_to_crystal_map(
        results, phase_dict, diffraction_library=diff_lib
    )
    assert xmap.shape == nav_shape
    assert np.allclose(xmap.phase_id, phase_id[:, 0])
    assert xmap.rotations_per_point == 1
    assert xmap.phases.names == phase_names
    assert (
        xmap.phases.structures[0].lattice.abcABG()
        == test_library_phases_multi.structures[0].lattice.abcABG()
    )

    # Raise warning when a property is not present as expected
    del results[prop_names[0]]
    with pytest.warns(UserWarning, match=f"Property '{prop_names[0]}' was expected"):
        xmap2 = results_dict_to_crystal_map(
            results, phase_dict, diffraction_library=diff_lib
        )
    assert list(xmap2.prop.keys()) == prop_names[1:]

    # Raise error when trying to access a best match which isn't
    # available
    with pytest.raises(ValueError, match="`index` cannot be higher than 2"):
        _ = results_dict_to_crystal_map(results, phase_dict, index=3)
    # Get second best match
    i = 1
    xmap3 = results_dict_to_crystal_map(results, phase_dict, index=i)
    assert xmap3.rotations_per_point == 1
    assert np.allclose(xmap3.phase_id, phase_id[:, i])
    assert np.allclose(xmap3.rotations.data, Rotation.from_euler(ori[:, i]).data)
    assert np.allclose(
        xmap3.prop[prop_names[1]], results[prop_names[1]][:, :, i].flatten()
    )

    # Make map single-phase and get all matches per point
    results["phase_index"][..., 0] = 0
    xmap4 = results_dict_to_crystal_map(results, phase_dict)
    assert np.all(xmap4.phase_id == 0)
    assert xmap4.phases.names[0] == phase_names[0]
    assert xmap4.rotations_per_point == 3

    # Get only best match even though map is single-phase
    i = 0
    xmap5 = results_dict_to_crystal_map(results, phase_dict, index=i)
    assert xmap5.rotations_per_point == 1
    assert np.allclose(xmap5.rotations.data, Rotation.from_euler(ori[:, i]).data)
    assert np.allclose(
        xmap5.prop[prop_names[1]], results[prop_names[1]][:, :, i].flatten()
    )



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.


import numpy as np
from warnings import warn

from scipy.spatial import distance_matrix
from sklearn.cluster import DBSCAN

from hyperspy.signals import Signal2D
from hyperspy._signals.lazy import LazySignal
from hyperspy.roi import CircleROI
from pyxem.signals.diffraction_vectors import DiffractionVectors, LazyDiffractionVectors
import hyperspy.api as hs


class DiffractionVectors2D(DiffractionVectors, Signal2D):
    """DiffractionVectors signal with the same number of vectors at each navigation
    position.

    This class can also be used to represent a list of vectors with no navigation axis.
    """

    _signal_dimension = 2
    _signal_type = "diffraction_vectors"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_unique_vectors(
        self,
        distance_threshold=0.01,
        method="distance_comparison",
        min_samples=1,
        return_clusters=False,
        columns=(-2, -1),
    ):
        """Returns diffraction vectors considered unique by:
        strict comparison, distance comparison with a specified
        threshold, or by clustering using DBSCAN [1].

        Parameters
        ----------
        distance_threshold : float
            The minimum distance between diffraction vectors for them to
            be considered unique diffraction vectors. If
            distance_threshold==0, the unique vectors will be determined
            by strict comparison.
        method : str
            The method to use to determine unique vectors. Valid methods
            are 'strict', 'distance_comparison' and 'DBSCAN'.
            'strict' returns all vectors that are strictly unique and
            corresponds to distance_threshold=0.
            'distance_comparison' checks the distance between vectors to
            determine if some should belong to the same unique vector,
            and if so, the unique vector is iteratively updated to the
            average value.
            'DBSCAN' relies on the DBSCAN [1] clustering algorithm, and
            uses the Eucledian distance metric.
        min_samples : int, optional
            The minimum number of not strictly identical vectors within
            one cluster for the cluster to be considered a core sample,
            i.e. to not be considered noise. Only used for method='DBSCAN'.
        return_clusters : bool, optional
            If True (False is default), the DBSCAN clustering result is
            returned. Only used for method='DBSCAN'.

        References
        ----------
        [1] "https://scikit-learn.org/stable/modules/generated/sklearn.cluster.DBSCAN.html"

        Returns
        -------
        unique_peaks : DiffractionVectors
            The unique diffraction vectors.
        clusters : DBSCAN
            The results from the clustering, given as class DBSCAN.
            Only returned if method='DBSCAN' and return_clusters=True.
        """

        # A distance_threshold of 0 implies a strict comparison. So in that
        # case, a warning is raised unless the specified method is 'strict'.
        data = self.data[:, columns]
        if distance_threshold == 0:
            if method != "strict":
                warn(
                    "distance_threshold=0 was given, and therefore "
                    "a strict comparison is used, even though the "
                    "specified method was {}".format(method)
                )
                method = "strict"

        if method == "strict":
            unique_peaks = np.unique(data, axis=0)

        elif method == "distance_comparison":
            unique_vectors, unique_counts = np.unique(data, axis=0, return_counts=True)

            unique_peaks = np.array([[0, 0]])
            unique_peaks_counts = np.array([0])

            while unique_vectors.shape[0] > 0:
                unique_vector = unique_vectors[0]
                distances = distance_matrix(np.array([unique_vector]), unique_vectors)
                indices = np.where(distances < distance_threshold)[1]

                new_count = indices.size
                new_unique_peak = np.array(
                    [
                        np.average(
                            unique_vectors[indices],
                            weights=unique_counts[indices],
                            axis=0,
                        )
                    ]
                )

                unique_peaks = np.append(unique_peaks, new_unique_peak, axis=0)

                unique_peaks_counts = np.append(unique_peaks_counts, new_count)
                unique_vectors = np.delete(unique_vectors, indices, axis=0)
                unique_counts = np.delete(unique_counts, indices, axis=0)
            unique_peaks = np.delete(unique_peaks, [0], axis=0)

        elif method == "DBSCAN":
            # All peaks are clustered by DBSCAN so that peaks within
            # one cluster are separated by distance_threshold or less.
            unique_vectors, unique_vectors_counts = np.unique(
                data, axis=0, return_counts=True
            )
            clusters = DBSCAN(
                eps=distance_threshold, min_samples=min_samples, metric="euclidean"
            ).fit(unique_vectors, sample_weight=unique_vectors_counts)
            unique_labels, unique_labels_count = np.unique(
                clusters.labels_, return_counts=True
            )
            unique_peaks = np.zeros((unique_labels.max() + 1, 2))

            # For each cluster, a center of mass is calculated based
            # on all the peaks within the cluster, and the center of
            # mass is taken as the final unique vector position.
            for n in np.arange(unique_labels.max() + 1):
                peaks_n_temp = unique_vectors[clusters.labels_ == n]
                peaks_n_counts_temp = unique_vectors_counts[clusters.labels_ == n]
                unique_peaks[n] = np.average(
                    peaks_n_temp, weights=peaks_n_counts_temp, axis=0
                )

        # Manipulate into DiffractionVectors class
        if unique_peaks.size > 0:
            unique_peaks = DiffractionVectors2D(unique_peaks)
        if return_clusters and method == "DBSCAN":
            return unique_peaks, clusters
        else:
            return unique_peaks

    def __lt__(self, other):
        return self._deepcopy_with_new_data(self.data < other)

    def __le__(self, other):
        return self._deepcopy_with_new_data(self.data <= other)

    def __gt__(self, other):
        return self._deepcopy_with_new_data(self.data > other)

    def __ge__(self, other):
        return self._deepcopy_with_new_data(self.data >= other)

    def from_peaks(cls, **kwargs):
        raise NotImplementedError("This method is not implemented for 2D vectors")

    def filter_detector_edge(self, exclude_width, columns=[-2, -1]):
        """Filter the diffraction vectors to accept only those not within a
        user specified proximity to the detector edge.

        Parameters
        ----------
        exclude_width : int
            The width of the region adjacent to the detector edge from which
            vectors will be excluded.
        Returns
        -------
        filtered_vectors : DiffractionVectors
            Diffraction vectors within allowed detector region.
        """
        values = self.data[:, columns]
        offsets = np.array(self.column_offsets)[columns]
        scales = np.array(self.column_scale)[columns]

        xlow = -offsets[0] + scales[0] * exclude_width
        xhigh = (
            -offsets[0]
            + (scales[0] * self.detector_shape[0])
            - (scales[0] * exclude_width)
        )

        ylow = -offsets[1] + scales[1] * exclude_width
        yhigh = (
            -offsets[1]
            + (scales[1] * self.detector_shape[1])
            - (scales[1] * exclude_width)
        )

        inbounds = (
            (values[:, 0] > xlow)
            * (values[:, 0] < xhigh)
            * (values[:, 1] > ylow)
            * (values[:, 1] < yhigh)
        )
        new_data = self.data[inbounds]

        return DiffractionVectors2D(new_data)

    # Functions that only work for 2D Vectors with no navigation axis

    @property
    def has_navigation_axis(self):
        return len(self.axes_manager.navigation_axes) > 0

    def to_roi(self, radius=0.1, columns=None, include_labels=False, **kwargs):
        """
        Convert the diffraction vectors to regions of interest (ROIs) for creating virtual images.

        Parameters
        ----------
        radius: float
            The radius of the ROIs in calibrated units
        columns: list
            The columns to use for the ROIs. If None, columns 0 and 1 are used to create the ROIs.
        include_labels: bool
            If True, the labels for each ROI are returned as well using hyperspy's
             :class:`hyperspy.api.plot.markers.Texts` class.
        kwargs: dict
            Keyword arguments to pass to the :class:`hyperspy.api.plot.markers.Texts` class.
        """
        if self.has_navigation_axis:
            raise NotImplementedError(
                "This method is not implemented for 2D vectors with a navigation axis"
            )
        if columns is None:
            columns = [0, 1]

        rois = [
            CircleROI(vector[columns[1]], vector[columns[0]], r=radius)
            for vector in self.data
        ]
        if include_labels:
            pos = [(vector[columns[1]], vector[columns[0]]) for vector in self.data]
            label = [f"{i}" for i in range(len(self.data))]
            texts = hs.plot.markers.Texts(offsets=pos, texts=label, **kwargs)
            return rois, texts
        else:
            return rois


class LazyDiffractionVectors2D(LazySignal, DiffractionVectors2D):
    pass



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

"""Generating subpixel resolution on diffraction vectors."""

import numpy as np
from skimage.registration import phase_cross_correlation
from skimage import draw
from pyxem.utils._deprecated import deprecated

from pyxem.signals import DiffractionVectors


def get_experimental_square(z, vector, square_size):
    """Defines a square region around a given diffraction vector and returns.

    Parameters
    ----------
    z : np.array()
        Single diffraction pattern
    vector : np.array()
        Single vector in pixels (int) [x,y] with top left as [0,0]
    square_size : int
        The length of one side of the bounding square (must be even)

    Returns
    -------
    square : np.array()
        Of size (L,L) where L = square_size

    """
    if square_size % 2 != 0:
        raise ValueError("'square_size' must be an even number")

    cx, cy, half_ss = vector[0], vector[1], int(square_size / 2)
    # select square with correct x,y see PR for details
    _z = z[cy - half_ss : cy + half_ss, cx - half_ss : cx + half_ss]
    return _z


def get_simulated_disc(square_size, disc_radius):
    """Create a uniform disc for correlating with the experimental square.

    Parameters
    ----------
    square size : int
        (even) - size of the bounding box
    disc_radius : int
        radius of the disc

    Returns
    -------
    arr: np.array()
        Upsampled copy of the simulated disc as a numpy array

    """

    if square_size % 2 != 0:
        raise ValueError("'square_size' must be an even number")

    ss = int(square_size)
    arr = np.zeros((ss, ss))
    rr, cc = draw.disk(
        (int(ss / 2), int(ss / 2)), radius=disc_radius, shape=arr.shape
    )  # is the thin disc a good idea
    arr[rr, cc] = 1
    return arr


def _get_pixel_vectors(dp, vectors, calibration, center):
    """Get the pixel coordinates for the given diffraction
    pattern and vectors.

    Parameters
    ----------
    dp: :obj:`pyxem.signals.ElectronDiffraction2D`
        Instance of ElectronDiffraction2D
    vectors : :obj:`pyxem.signals.diffraction_vectors.DiffractionVectors`
        List of diffraction vectors
    calibration : [float, float]
        Calibration values
    center : float, float
        Image origin in pixel coordinates

    Returns
    -------
    vector_pixels : :obj:`pyxem.signals.diffraction_vectors.DiffractionVectors`
        Pixel coordinates for given diffraction pattern and vectors.
    """

    def _floor(vectors, calibration, center):
        if vectors.shape == (1,) and vectors.dtype == object:
            vectors = vectors[0]
        return np.floor((vectors.astype(np.float64) / calibration) + center).astype(int)

    if isinstance(vectors, DiffractionVectors):
        if vectors.axes_manager.navigation_shape != dp.axes_manager.navigation_shape:
            raise ValueError(
                "Vectors with shape {} must have the same navigation shape "
                "as the diffraction patterns which has shape {}.".format(
                    vectors.axes_manager.navigation_shape,
                    dp.axes_manager.navigation_shape,
                )
            )
        vector_pixels = vectors.map(
            _floor, calibration=calibration, center=center, inplace=False
        )
    else:
        vector_pixels = _floor(vectors, calibration, center)

    dp_max_size = np.max(dp.data.shape) - 1
    if isinstance(vector_pixels, DiffractionVectors):
        min_value, max_value = 100, -100
        for index in np.ndindex(dp.axes_manager.navigation_shape):
            islice = np.s_[index]
            vec = vector_pixels.inav[islice].data[0]
            if len(vec) == 0:
                temp_min = 0
                temp_max = 0
            else:
                temp_min = vec.flatten().min()
                temp_max = vec.flatten().max()
            if min_value > temp_min:
                min_value = temp_min
            if max_value < temp_max:
                max_value = temp_max
        if max_value > dp_max_size or (min_value < 0):
            raise ValueError(
                "Some of your vectors do not lie within your "
                "diffraction pattern, check your calibration"
            )
    elif isinstance(vector_pixels, np.ndarray):
        if np.any(vector_pixels > dp_max_size) or np.any(vector_pixels < 0):
            raise ValueError(
                "Some of your vectors do not lie within your "
                "diffraction pattern, check your calibration"
            )
    else:
        raise ValueError(
            "The vectors are not recognized as either DiffractionVectors "
            "or a NumPy array"
        )
    return vector_pixels


def _conventional_xc(exp_disc, sim_disc, upsample_factor):
    """Takes two images of disc and finds the shift between them using
    conventional (phase) cross correlation.

    Parameters
    ----------
    exp_disc : np.array()
        A numpy array of the "experimental" disc
    sim_disc : np.array()
        A numpy array of the disc used as a template
    upsample_factor: int (must be even)
        Factor to upsample by, reciprocal of the subpixel resolution
        (eg 10 ==> 1/10th of a pixel)

    Returns
    -------
    shifts
        Pixel shifts required to register the two images

    """

    shifts, error, _ = phase_cross_correlation(
        exp_disc, sim_disc, upsample_factor=upsample_factor
    )
    shifts = np.flip(shifts)  # to comply with hyperspy conventions - see issue#490
    return shifts


def _center_of_mass_hs(z):
    """Return the center of mass of an array with coordinates in the
    hyperspy convention

    Parameters
    ----------
    z : np.array

    Returns
    -------
    (x,y) : tuple of floats
        The x and y locations of the center of mass of the parsed square
    """

    s = np.sum(z)
    if s != 0:
        z *= 1 / s
    dx = np.sum(z, axis=0)
    dy = np.sum(z, axis=1)
    h, w = z.shape
    cx = np.sum(dx * np.arange(w))
    cy = np.sum(dy * np.arange(h))
    return cx, cy


def _com_experimental_square(z, vector, square_size):
    """Wrapper for get_experimental_square that makes the non-zero
    elements symmetrical around the 'unsubpixeled' peak by zeroing a
    'spare' row and column (top and left).

    Parameters
    ----------
    z : np.array

    vector : np.array([x,y])

    square_size : int (even)

    Returns
    -------
    z_adpt : np.array
        z, but with row and column zero set to 0
    """
    # Copy to make sure we don't change the dp
    z_adpt = np.copy(get_experimental_square(z, vector=vector, square_size=square_size))
    z_adpt[:, 0] = 0
    z_adpt[0, :] = 0
    return z_adpt


def _center_of_mass_map(dp, vectors, square_size, center, calibration, columns=None):
    if vectors.shape == (2,):
        vectors = np.array(
            [
                vectors,
            ]
        )
    shifts = np.zeros_like(vectors, dtype=np.float64)
    if columns is not None:
        num_cols = vectors.shape[1]
        other_indexes = np.arange(num_cols)[
            np.logical_not(np.isin(np.arange(num_cols), columns))
        ]
        vectors = vectors[:, columns]
        extra_columns = vectors[:, other_indexes]

    for i, vector in enumerate(vectors):
        expt_disc = _com_experimental_square(dp, vector, square_size)
        shifts[i] = [a - square_size / 2 for a in _center_of_mass_hs(expt_disc)]
    new_vectors = ((vectors + shifts) - center) * calibration
    if columns is not None:
        new_vectors = np.hstack(new_vectors, extra_columns)
    return new_vectors


class SubpixelrefinementGenerator:
    """Generates subpixel refinement of DiffractionVectors.

    Parameters
    ----------
    dp : ElectronDiffraction2D
        The electron diffraction patterns to be refined
    vectors : DiffractionVectors | ndarray
        Vectors (in calibrated units) to the locations of the spots to be
        refined. If given as DiffractionVectors, it must have the same
        navigation shape as the electron diffraction patterns. If an ndarray,
        the same set of vectors is mapped over all electron diffraction
        patterns.

    References
    ----------
    [1] Pekin et al. Ultramicroscopy 176 (2017) 170-176

    """

    @deprecated(
        since="0.17.0",
        removal="1.0.0",
        alternative="pyxem.signals.DiffractionVectors.subpixel_refine",
    )
    def __init__(self, dp, vectors):
        self.dp = dp
        self.vectors_init = vectors
        self.last_method = None
        sig_ax = dp.axes_manager.signal_axes
        self.calibration = [sig_ax[0].scale, sig_ax[1].scale]
        self.center = [sig_ax[0].size / 2, sig_ax[1].size / 2]

        self.vector_pixels = _get_pixel_vectors(
            dp, vectors, calibration=self.calibration, center=self.center
        )

    def conventional_xc(self, square_size, disc_radius, upsample_factor):
        """Refines the peaks using (phase) cross correlation.

        Parameters
        ----------
        square_size : int
            Length (in pixels) of one side of a square the contains the peak to
            be refined.
        disc_radius:  int
            Radius (in pixels) of the discs that you seek to refine
        upsample_factor: int
            Factor by which to upsample the patterns

        Returns
        -------
        vector_out: DiffractionVectors
            DiffractionVectors containing the refined vectors in calibrated
            units with the same navigation shape as the diffraction patterns.

        """

        def _conventional_xc_map(
            dp, vectors, sim_disc, upsample_factor, center, calibration
        ):
            shifts = np.zeros_like(vectors, dtype=np.float64)
            for i, vector in enumerate(vectors):
                expt_disc = get_experimental_square(dp, vector, square_size)
                shifts[i] = _conventional_xc(expt_disc, sim_disc, upsample_factor)
            return ((vectors + shifts) - center) * calibration

        sim_disc = get_simulated_disc(square_size, disc_radius)
        self.vectors_out = self.dp.map(
            _conventional_xc_map,
            vectors=self.vector_pixels,
            sim_disc=sim_disc,
            upsample_factor=upsample_factor,
            center=self.center,
            calibration=self.calibration,
            inplace=False,
            ragged=True,
        )
        self.vectors_out.set_signal_type("diffraction_vectors")

        self.last_method = "conventional_xc"
        return self.vectors_out

    def reference_xc(self, square_size, reference_dp, upsample_factor):
        """Refines the peaks using (phase) cross correlation with a reference
        diffraction image.

        Parameters
        ----------
        square_size : int
            Length (in pixels) of one side of a square the contains the peak to
            be refined.
        reference_dp: ndarray
            Same shape as a single diffraction image
        upsample_factor: int
            Factor by which to upsample the patterns

        Returns
        -------
        vector_out: DiffractionVectors
            DiffractionVectors containing the refined vectors in calibrated
            units with the same navigation shape as the diffraction patterns.

        """

        def _reference_xc_map(dp, vectors, upsample_factor, center, calibration):
            shifts = np.zeros_like(vectors, dtype=np.float64)
            for i, vector in enumerate(vectors):
                ref_disc = get_experimental_square(reference_dp, vector, square_size)
                expt_disc = get_experimental_square(dp, vector, square_size)
                shifts[i] = _conventional_xc(expt_disc, ref_disc, upsample_factor)
            return ((vectors + shifts) - center) * calibration

        self.vectors_out = self.dp.map(
            _reference_xc_map,
            vectors=self.vector_pixels,
            upsample_factor=upsample_factor,
            center=self.center,
            calibration=self.calibration,
            inplace=False,
            ragged=True,
        )
        self.vectors_out.set_signal_type("diffraction_vectors")

        self.last_method = "reference_xc"
        return self.vectors_out

    def center_of_mass_method(self, square_size):
        """Find the subpixel refinement of a peak by assuming it lies at the
        center of intensity.

        Parameters
        ----------
        square_size : int
            Length (in pixels) of one side of a square the contains the peak to
            be refined.

        Returns
        -------
        vector_out: DiffractionVectors
            DiffractionVectors containing the refined vectors in calibrated
            units with the same navigation shape as the diffraction patterns.

        """
        self.vectors_out = self.dp.map(
            _center_of_mass_map,
            vectors=self.vector_pixels,
            square_size=square_size,
            center=self.center,
            calibration=self.calibration,
            inplace=False,
            ragged=True,
        )
        self.vectors_out.set_signal_type("diffraction_vectors")
        if not isinstance(self.vectors_init, DiffractionVectors):
            self.vectors_out.scales = 1
            self.vectors_out.offsets = 0
        else:
            self.vectors_out.scales = self.vectors_init.scales
            self.vectors_out.offsets = self.vectors_init.offsets
            self.vectors_out.detector_shape = self.vectors_init.detector_shape

        self.last_method = "center_of_mass_method"
        return self.vectors_out

    def local_gaussian_method(self, square_size):
        """Removed in v0.13, please install a version prior to v.0.13 to use."""
        raise NotImplementedError(
            "This functionality was removed in v.0.13.0, please use another method"
        )



# -*- coding: utf-8 -*-
# Copyright 2016-2023 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

"""Utils for subpixel vectors refinement."""

import numpy as np
from skimage.transform import rescale
from skimage import draw

from pyxem.utils.diffraction import normalize_template_match


def _get_experimental_square(z, vector, square_size):
    """Defines a square region around a given diffraction vector and returns.

    Parameters
    ----------
    z : np.array()
        Single diffraction pattern
    vector : np.array()
        Single vector in pixels (int) [x,y] with top left as [0,0]
    square_size : int
        The length of one side of the bounding square (must be even)

    Returns
    -------
    square : np.array()
        Of size (L,L) where L = square_size

    """
    if square_size % 2 != 0:
        raise ValueError("'square_size' must be an even number")
    # reverse the order of the vector
    cx, cy, half_ss = int(vector[1]), int(vector[0]), int(square_size / 2)
    _z = z[cy - half_ss : cy + half_ss + 1, cx - half_ss : cx + half_ss + 1]
    return _z


def _get_simulated_disc(square_size, disc_radius):
    """Create a uniform disc for correlating with the experimental square.

    Parameters
    ----------
    square size : int
        (even) - size of the bounding box
    disc_radius : int
        radius of the disc

    Returns
    -------
    arr: np.array()
        Upsampled copy of the simulated disc as a numpy array

    """

    if square_size % 2 != 0:
        raise ValueError("'square_size' must be an even number")

    ss = int(square_size)
    arr = np.zeros((ss, ss))
    rr, cc = draw.disk(
        (int(ss / 2), int(ss / 2)), radius=disc_radius, shape=arr.shape
    )  # is the thin disc a good idea
    arr[rr, cc] = 1
    return arr


def _center_of_mass_hs(z):
    """Return the center of mass of an array with coordinates in the
    hyperspy convention

    Parameters
    ----------
    z : np.array

    Returns
    -------
    (x,y) : tuple of floats
        The x and y locations of the center of mass of the parsed square
    """

    s = np.sum(z)
    if s != 0:
        z *= 1 / s
    dx = np.sum(z, axis=0)
    dy = np.sum(z, axis=1)
    h, w = z.shape
    cx = np.sum(dx * np.arange(w))
    cy = np.sum(dy * np.arange(h))
    return cy, cx


def _com_experimental_square(z, vector, square_size):
    """Wrapper for get_experimental_square that makes the non-zero
    elements symmetrical around the 'unsubpixeled' peak by zeroing a
    'spare' row and column (top and left).

    Parameters
    ----------
    z : np.array

    vector : np.array([x,y])

    square_size : int (even)

    Returns
    -------
    z_adpt : np.array
        z, but with row and column zero set to 0
    """
    # Copy to make sure we don't change the dp
    z_adpt = np.copy(
        _get_experimental_square(z, vector=vector, square_size=square_size)
    )
    z_adpt[:, 0] = 0
    z_adpt[0, :] = 0
    return z_adpt


def _conventional_xc(slic, kernel, upsample_factor):
    """Takes two images of disc and finds the shift between them using
    conventional cross correlation.
    """
    half_ss = int((slic.shape[0]) / 2)
    slic = rescale(slic, upsample_factor, order=1, mode="reflect")
    kernel = rescale(kernel, upsample_factor, order=1, mode="reflect")

    temp = normalize_template_match(slic, kernel)
    max = np.array(np.unravel_index(np.argmax(temp), temp.shape))
    shifts = np.array(max) / upsample_factor - half_ss
    return shifts


#####################################################
# Methods for subpixel refinement on a set of vectors
#####################################################


def _center_of_mass_map(dp, vectors, square_size, offsets, scales):
    shifts = np.zeros_like(vectors, dtype=np.float64)
    for i, vector in enumerate(vectors):
        square = _get_experimental_square(dp, vector, square_size)
        shifts[i] = [a - square_size / 2 for a in _center_of_mass_hs(square)]

    new_vectors = (vectors + shifts) * scales + offsets
    return new_vectors


def _conventional_xc_map(
    dp,
    vectors,
    kernel,
    square_size,
    upsample_factor,
    offsets,
    scales,
):
    vectors = np.array(vectors).astype(int)
    shifts = np.zeros_like(vectors, dtype=np.float64)
    for i, vector in enumerate(vectors):
        expt_disc = _get_experimental_square(dp, vector, square_size)
        shifts[i] = _conventional_xc(expt_disc, kernel, upsample_factor)

    return (vectors + shifts) * scales + offsets


def _wrap_columns(dp, vectors, f, columns=None, **kwargs):
    """
    Take some function, f, and apply it to the 2d array X returning a list
    of peak positions.  The intensity at each peak position is then appended
    to the returned list of peaks

    Parameters
    ----------
    X: 2-D array-like
        The input image used to find peaks
    f: func
        The passed function to find peaks
    kwargs:
        Any additional keyword arguments passed to f

    Returns
    -------
        peaks:array-like
            A 2d array with columns [x, y, intensity]

    """
    if vectors.shape == (2,):
        vectors = np.array(
            [
                vectors,
            ]
        )
    if columns is not None:
        num_cols = vectors.shape[1]
        other_indexes = np.arange(num_cols)[
            np.logical_not(np.isin(np.arange(num_cols), columns))
        ]
        extra_columns = vectors[:, other_indexes]
        vectors = vectors[:, columns]
    new_vectors = f(dp, vectors[:, columns], **kwargs)
    if columns is not None:
        new_vectors = np.hstack((new_vectors, extra_columns))
    return new_vectors



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

"""Signal classes for diffraction data.
"""


from .common_diffraction import CommonDiffraction
from .correlation2d import Correlation2D, LazyCorrelation2D
from .diffraction_variance1d import DiffractionVariance1D
from .diffraction_variance2d import DiffractionVariance2D, ImageVariance
from .diffraction_vectors import DiffractionVectors
from .diffraction_vectors2d import DiffractionVectors2D
from .diffraction_vectors1d import DiffractionVectors1D
from .beam_shift import BeamShift, LazyBeamShift
from .differential_phase_contrast import (
    DPCSignal1D,
    DPCSignal2D,
    LazyDPCSignal1D,
    LazyDPCSignal2D,
)
from .diffraction1d import Diffraction1D, LazyDiffraction1D
from .diffraction2d import Diffraction2D, LazyDiffraction2D
from .polar_vectors import PolarVectors, LazyPolarVectors
from .electron_diffraction1d import ElectronDiffraction1D, LazyElectronDiffraction1D
from .electron_diffraction2d import ElectronDiffraction2D, LazyElectronDiffraction2D
from .indexation_results import VectorMatchingResults, OrientationMap
from .pair_distribution_function1d import PairDistributionFunction1D
from .polar_diffraction2d import PolarDiffraction2D, LazyPolarDiffraction2D
from .power2d import Power2D, LazyPower2D
from .reduced_intensity1d import ReducedIntensity1D
from .segments import LearningSegment, VDFSegment
from .strain_map import StrainMap
from .correlation1d import Correlation1D, LazyCorrelation1D
from .tensor_field import DisplacementGradientMap
from .virtual_dark_field_image import VirtualDarkFieldImage
from .insitu_diffraction2d import InSituDiffraction2D
from .labeled_diffraction_vectors2d import LabeledDiffractionVectors2D


__all__ = [
    "CommonDiffraction",
    "Correlation2D",
    "LazyCorrelation2D",
    "BeamShift",
    "LazyBeamShift",
    "DPCBaseSignal",
    "DPCSignal1D",
    "DPCSignal2D",
    "LazyDPCBaseSignal",
    "LazyDPCSignal1D",
    "LazyDPCSignal2D",
    "DiffractionVariance1D",
    "DiffractionVariance2D",
    "ImageVariance",
    "DiffractionVectors",
    "DiffractionVectors1D",
    "DiffractionVectors2D",
    "Diffraction1D",
    "LazyDiffraction1D",
    "LazyDiffraction2D",
    "Diffraction2D",
    "ElectronDiffraction1D",
    "LazyElectronDiffraction1D",
    "LazyElectronDiffraction2D",
    "ElectronDiffraction2D",
    "VectorMatchingResults",
    "PairDistributionFunction1D",
    "PolarDiffraction2D",
    "LazyPolarDiffraction2D",
    "PolarVectors",
    "LazyPolarVectors",
    "Power2D",
    "LazyPower2D",
    "ReducedIntensity1D",
    "LearningSegment",
    "VDFSegment",
    "StrainMap",
    "Correlation1D",
    "LazyCorrelation1D",
    "DisplacementGradientMap",
    "VirtualDarkFieldImage",
    "InSituDiffraction2D",
    "LabeledDiffractionVectors2D",
    "OrientationMap",
]



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from tqdm import tqdm
from scipy.ndimage import rotate, gaussian_filter
from scipy.signal import convolve2d
from skimage import morphology
import dask.array as da

from hyperspy.misc.utils import isiterable

from pyxem.signals import Diffraction2D, LazyDiffraction2D
import pyxem.utils.ransac_ellipse_tools as ret


def _get_elliptical_disk(xx, yy, x, y, semi_len0, semi_len1, rotation):
    """Private function to generate an elliptical disk.

    Parameters
    ----------
    xx, yy : 2D NumPy array
    x, y : float
    semi_len0, semi_len1 : float
    rotation : float
        In radians

    Returns
    -------
    ellipse_array : 2D NumPy array

    Examples
    --------
    >>> from hyperspy.signals import Signal2D
    >>> from pyxem.data.dummy_data import make_diffraction_test_data as mdtd
    >>> s = Signal2D(np.zeros((110, 130)))
    >>> s.axes_manager[0].offset, s.axes_manager[1].offset = -50, -80
    >>> xx, yy = np.meshgrid(
    ...     s.axes_manager.signal_axes[0].axis,
    ...     s.axes_manager.signal_axes[1].axis)
    >>> ellipse_data = mdtd._get_elliptical_disk(xx, yy, 10, -10, 12, 18, 1.5)
    >>> s.data += ellipse_data
    >>> s.plot()

    """
    rot = rotation
    xx0 = xx - x
    yy0 = yy - y
    semi_len02 = semi_len0 * semi_len0
    semi_len12 = semi_len1 * semi_len1
    z0 = ((xx0 * np.cos(rot) + yy0 * np.sin(rot)) ** 2) / semi_len02
    z1 = ((xx0 * np.sin(rot) - yy0 * np.cos(rot)) ** 2) / semi_len12
    zz = z0 + z1
    elli_mask = zz <= 1.0
    return elli_mask


def _get_elliptical_ring(xx, yy, x, y, semi_len0, semi_len1, rotation, lw_r=1):
    """Private function to generate an elliptical ring.

    Parameters
    ----------
    xx, yy : 2D NumPy array
    x, y : float
    semi_len0, semi_len1 : float
    rotation : float
        In radians
    lw_r : optional, default 1

    Returns
    -------
    ellipse_array : 2D NumPy array

    Examples
    --------
    >>> from hyperspy.signals import Signal2D
    >>> from pyxem.data.dummy_data import make_diffraction_test_data as mdtd
    >>> s = Signal2D(np.zeros((110, 130)))
    >>> s.axes_manager[0].offset, s.axes_manager[1].offset = -50, -80
    >>> xx, yy = np.meshgrid(
    ...     s.axes_manager.signal_axes[0].axis,
    ...     s.axes_manager.signal_axes[1].axis)
    >>> ellipse_data = mdtd._get_elliptical_ring(
    ...     xx, yy, 10, -10, 12, 18, 1.5, 2)
    >>> s.data += ellipse_data
    >>> s.plot()

    """
    mask_outer = _get_elliptical_disk(
        xx, yy, x, y, semi_len0 + lw_r, semi_len1 + lw_r, rotation
    )
    mask_inner = _get_elliptical_disk(
        xx, yy, x, y, semi_len0 - lw_r, semi_len1 - lw_r, rotation
    )
    ellipse = np.logical_xor(mask_outer, mask_inner)
    ellipse = ellipse.astype("uint32")
    return ellipse


class EllipseRing:
    """Generate an elliptical ring.

    Parameters
    ----------
    xx, yy : int
        Size of the image
    x, y : float
        Centre positions for the ellipse
    semi_len0, semi_len1 : float
    rotation : float
        In radians
    lw_r : optional, default 1

    Methods
    -------
    get_signal
        Generate numpy array with the ellipse ring

    Examples
    --------
    >>> from pyxem.data.dummy_data import make_diffraction_test_data as mdtd
    >>> ellipse = mdtd.EllipseRing(
    ...     xx=20, yy=30, x0=10, y0=15, semi_len0=4, semi_len1=6,
    ...     rotation=1.57, intensity=10, lw_r=2)

    """

    def __init__(
        self, xx, yy, x0, y0, semi_len0, semi_len1, rotation, intensity, lw_r=1
    ):
        self.x0, self.y0 = x0, y0
        self.semi_len0 = semi_len0
        self.semi_len1 = semi_len1
        self.rotation = rotation
        self.intensity = intensity
        self.xx, self.yy = xx, yy
        self.lw_r = lw_r

    def __repr__(self):
        return (
            "<%s, ((x0, y0): (%s, %s), (sl0, sl1): (%s, %s),"
            " r: %s, I: %s, lw: %s)>"
            % (
                self.__class__.__name__,
                self.x0,
                self.y0,
                self.semi_len0,
                self.semi_len1,
                self.rotation,
                self.intensity,
                self.lw_r,
            )
        )

    def get_signal(self):
        ellipse = _get_elliptical_ring(
            xx=self.xx,
            yy=self.yy,
            x=self.x0,
            y=self.y0,
            semi_len0=self.semi_len0,
            semi_len1=self.semi_len1,
            rotation=self.rotation,
            lw_r=self.lw_r,
        )
        self.ellipse = ellipse * self.intensity
        return self.ellipse


class EllipseDisk:
    """Generate an elliptical disk.

    Parameters
    ----------
    xx, yy : int
        Size of the image
    x, y : float
        Centre positions for the ellipse
    semi_len0, semi_len1 : float
    rotation : float
        In radians
    lw_r : optional, default 1

    Methods
    -------
    get_signal
        Generate numpy array with the elliptical disk

    Examples
    --------
    >>> from pyxem.data.dummy_data import make_diffraction_test_data as mdtd
    >>> ellipse_disk = mdtd.EllipseDisk(
    ...     xx=20, yy=30, x0=10, y0=15, semi_len0=4, semi_len1=6,
    ...     rotation=1.57, intensity=10)

    """

    def __init__(self, xx, yy, x0, y0, semi_len0, semi_len1, rotation, intensity):
        self.x0, self.y0 = x0, y0
        self.semi_len0 = semi_len0
        self.semi_len1 = semi_len1
        self.rotation = rotation
        self.intensity = intensity
        self.xx, self.yy = xx, yy

    def __repr__(self):
        return "<%s, ((x0, y0): (%s, %s), (sl0, sl1): (%s, %s)," " r: %s, I: %s)>" % (
            self.__class__.__name__,
            self.x0,
            self.y0,
            self.semi_len0,
            self.semi_len1,
            self.rotation,
            self.intensity,
        )

    def get_signal(self):
        ellipse = _get_elliptical_disk(
            xx=self.xx,
            yy=self.yy,
            x=self.x0,
            y=self.y0,
            semi_len0=self.semi_len0,
            semi_len1=self.semi_len1,
            rotation=self.rotation,
        )
        self.ellipse = ellipse * self.intensity
        return self.ellipse


class Circle:
    def __init__(self, xx, yy, x0, y0, r, intensity, scale, lw=None):
        self.x0 = x0
        self.y0 = y0
        self.r = r
        self.intensity = intensity
        self.lw = lw
        self.circle = (yy - self.y0) ** 2 + (xx - self.x0) ** 2
        self.mask_outside_r(scale)
        self.get_centre_pixel(xx, yy, scale)

    def __repr__(self):
        return "<%s, (r: %s, (x0, y0): (%s, %s), I: %s)>" % (
            self.__class__.__name__,
            self.r,
            self.x0,
            self.y0,
            self.intensity,
        )

    def mask_outside_r(self, scale):
        if self.lw is None:
            indices = self.circle >= (self.r + scale) ** 2
        else:
            indices = self.circle >= (self.r + self.lw + scale) ** 2
        self.circle[indices] = 0

    def centre_on_image(self, xx, yy):
        if self.x0 < xx[0][0] or self.x0 > xx[0][-1]:
            return False
        elif self.y0 < yy[0][0] or self.y0 > yy[-1][-1]:
            return False
        else:
            return True

    def get_centre_pixel(self, xx, yy, scale):
        """Sets the indices for the pixels on which the centre point is.

        Because the centre point can sometimes be exactly on the
        boundary of two pixels, the pixels are held in a list. One
        list for x (self.centre_x_pixels) and one for y
        (self.centre_x_pixels). If the centre is outside the image, the
        lists will be empty.
        """
        if self.centre_on_image(xx, yy):
            x1 = np.where(xx > (self.x0 - 0.5 * scale))[1][0]
            x2 = np.where(xx < (self.x0 + 0.5 * scale))[1][-1]
            self.centre_x_pixels = [x1, x2]
            y1 = np.where(yy > (self.y0 - 0.5 * scale))[0][0]
            y2 = np.where(yy < (self.y0 + 0.5 * scale))[0][-1]
            self.centre_y_pixels = [y1, y2]
        else:
            self.centre_x_pixels = []
            self.centre_y_pixels = []

    def set_uniform_intensity(self):
        circle_ring_indices = self.circle > 0
        self.circle[circle_ring_indices] = self.intensity


class Disk:
    """Disk object, with outer edge of the ring at r."""

    def __init__(self, xx, yy, scale, x0, y0, r, intensity):
        self.z = Circle(xx, yy, x0, y0, r, intensity, scale)
        self.z.set_uniform_intensity()
        self.set_centre_intensity()

    def __repr__(self):
        return "<%s, (r: %s, (x0, y0): (%s, %s), I: %s)>" % (
            self.__class__.__name__,
            self.z.r,
            self.z.x0,
            self.z.y0,
            self.z.intensity,
        )

    def set_centre_intensity(self):
        """Sets the intensity of the centre pixels to I.

        Coordinates are self.z.circle[y, x], due to how numpy works.
        """
        for x in self.z.centre_x_pixels:
            for y in self.z.centre_y_pixels:
                self.z.circle[y, x] = self.z.intensity  # This is correct

    def get_signal(self):
        return self.z.circle


class Ring:
    """Ring object, with outer edge of the ring at r+lr, and inner r-lr.

    The radius of the ring is defined as in the middle of the line making
    up the ring.

    """

    def __init__(self, xx, yy, scale, x0, y0, r, intensity, lr):
        if lr > r:
            raise ValueError(f"Ring line width too big ({lr} > {r})")
        self.lr = lr
        self.lw = 1 + 2 * lr  # scalar line width of the ring
        self.z = Circle(xx, yy, x0, y0, r, intensity, scale, lw=lr)
        self.mask_inside_r(scale)
        self.z.set_uniform_intensity()

    def __repr__(self):
        return "<%s, (r: %s, (x0, y0): (%s, %s), I: %s)>" % (
            self.__class__.__name__,
            self.z.r,
            self.z.x0,
            self.z.y0,
            self.z.intensity,
        )

    def mask_inside_r(self, scale):
        indices = self.z.circle < (self.z.r - self.lr) ** 2
        self.z.circle[indices] = 0

    def get_signal(self):
        return self.z.circle


class MakeTestData:
    """MakeTestData is an object containing a generated test signal.

    The default signal consists of a Disk and concentric Ring, with the
    Ring being less intensive than the centre Disk. Unlimited number of
    Rings and Disk can be added separately.

    Parameters
    ----------
    size_x, size_y : float, int
        The range of the x and y axis goes from 0 to size_x, size_y
    scale : float, int
        The step size of the x and y axis
    default : bool, default False
        If True, the default object should be generated. If false, Ring and
        Disk must be added separately by self.add_ring(), self.add_disk()
    blur : bool, default True
        If True, do a Gaussian blur of the disk.
    blur_sigma : int, default 1
        Sigma of the Gaussian blurring, if blur is True.
    downscale : bool, default True
        Note: currently using downscale and adding a disk, will lead to the
        center of the disk being shifted. Ergo: the disk_x and disk_y will not
        be correct when downscale is True.

    Attributes
    ----------
    signal : hyperspy.signals.Signal2D
        Test signal
    z_list : list
        List containing Ring and Disk objects added to the signal
    downscale_factor : int
        The data is upscaled before Circle is added, and similarly
        downscaled to return to given dimensions. This improves the
        quality of Circle

    Examples
    --------
    Default settings

    >>> from pyxem.data.dummy_data.make_diffraction_test_data import MakeTestData
    >>> test_data = MakeTestData()
    >>> test_data.signal.plot()

    More control

    >>> test_data = MakeTestData(default=False)
    >>> test_data.add_disk(x0=50, y0=50, r=10, intensity=30)
    >>> test_data.add_ring(x0=45, y0=52, r=25, intensity=10)
    >>> test_data.signal.plot()

    """

    def __init__(
        self,
        size_x=100,
        size_y=100,
        scale=1,
        default=False,
        blur=True,
        blur_sigma=1,
        downscale=True,
    ):
        self.scale = scale
        self.blur_on = blur
        self.blur_sigma = blur_sigma
        self.downscale_on = downscale
        self.downscale_factor = 5
        if not self.downscale_on:
            self.downscale_factor = 1
        self.size_x, self.size_y = size_x, size_y
        self.generate_mesh()
        self.z_list = []
        if default:
            self.add_disk()
            self.add_ring(lw_pix=1)
        else:
            self.update_signal()

    def __repr__(self):
        return "<%s, ((x, y): (%s, %s), s: %s, z: %s)>" % (
            self.__class__.__name__,
            self.size_x,
            self.size_y,
            self.scale,
            len(self.z_list),
        )

    def update_signal(self):
        self.make_signal()
        self.downscale()
        self.blur()
        self.to_signal()

    def generate_mesh(self):
        self.X = np.arange(0, self.size_x, self.scale / self.downscale_factor)
        self.Y = np.arange(0, self.size_y, self.scale / self.downscale_factor)
        self.xx, self.yy = np.meshgrid(self.X, self.Y, sparse=True)

    def add_disk(self, x0=50, y0=50, r=5, intensity=10):
        scale = self.scale / self.downscale_factor
        self.z_list.append(Disk(self.xx, self.yy, scale, x0, y0, r, intensity))
        self.update_signal()

    def add_disk_ellipse(
        self, x0=50, y0=50, semi_len0=5, semi_len1=8, rotation=0.78, intensity=10
    ):
        ellipse = EllipseDisk(
            xx=self.xx,
            yy=self.yy,
            x0=x0,
            y0=y0,
            semi_len0=semi_len0,
            semi_len1=semi_len1,
            rotation=rotation,
            intensity=intensity,
        )
        self.z_list.append(ellipse)
        self.update_signal()

    def add_ring(self, x0=50, y0=50, r=20, intensity=10, lw_pix=0):
        """
        Add a ring to the test data.

        Parameters
        ----------
        x0, y0 : number, default 50
            Centre position of the ring
        r : number, default 20
            Radius of the ring, defined as the distance from the centre to the
            middle of the line of the ring, which will be most intense after
            blurring.
        intensity : number, default 10
            Pixel value of the ring. Note, this value will be lowered
            if blur or downscale is True
        lw_pix : number, default 0
            Distance in pixels from radius to the outer and inner edge of the
            ring. Inner radius: r-lw, outer radius: r+lw. In total this gives
            a ring line width in pixels of 2*lw+1.

        """
        scale = self.scale / self.downscale_factor
        lr = lw_pix * self.scale  # scalar
        self.z_list.append(Ring(self.xx, self.yy, scale, x0, y0, r, intensity, lr))
        self.update_signal()

    def add_ring_ellipse(
        self,
        x0=50,
        y0=50,
        semi_len0=11,
        semi_len1=13,
        rotation=0.78,
        intensity=10,
        lw_r=2,
    ):
        ellipse = EllipseRing(
            xx=self.xx,
            yy=self.yy,
            x0=x0,
            y0=y0,
            semi_len0=semi_len0,
            semi_len1=semi_len1,
            rotation=rotation,
            intensity=intensity,
            lw_r=lw_r,
        )
        self.z_list.append(ellipse)
        self.update_signal()

    def make_signal(self):
        if len(self.z_list) == 0:
            self.z = self.xx * 0 + self.yy * 0
        elif len(self.z_list) == 1:
            self.z = self.z_list[0].get_signal()
        elif len(self.z_list) > 1:
            z_temp = self.z_list[0].get_signal()
            for i in self.z_list[1:]:
                z_temp = np.add(z_temp, i.get_signal())
            self.z = z_temp

    def downscale(self):
        if self.downscale_on:
            shape = (
                int(self.z.shape[0] / self.downscale_factor),
                int(self.z.shape[1] / self.downscale_factor),
            )
            sh = (
                shape[0],
                self.z.shape[0] // shape[0],
                shape[1],
                self.z.shape[1] // shape[1],
            )
            self.z_downscaled = self.z.reshape(sh).mean(-1).mean(1)
        else:
            self.z_downscaled = self.z

    def blur(self):
        if self.blur_on:
            self.z_blurred = gaussian_filter(self.z_downscaled, sigma=self.blur_sigma)
        else:
            self.z_blurred = self.z_downscaled

    def to_signal(self):
        self.signal = Diffraction2D(self.z_blurred)
        self.signal.axes_manager[0].scale = self.scale
        self.signal.axes_manager[1].scale = self.scale

    def set_signal_zero(self):
        self.z_list = []
        self.update_signal()


def generate_4d_data(
    probe_size_x=10,
    probe_size_y=10,
    image_size_x=50,
    image_size_y=50,
    disk_x=25,
    disk_y=25,
    disk_r=5,
    disk_I=20,
    ring_x=25,
    ring_y=25,
    ring_r=20,
    ring_I=6,
    ring_lw=0,
    ring_e_x=None,
    ring_e_y=25,
    ring_e_semi_len0=15,
    ring_e_semi_len1=15,
    ring_e_r=0,
    ring_e_I=6,
    ring_e_lw=1,
    blur=True,
    blur_sigma=1,
    downscale=True,
    add_noise=False,
    noise_amplitude=1,
    lazy=False,
    lazy_chunks=None,
    show_progressbar=True,
):
    """Generate a test dataset containing a disk and diffraction ring.

    Useful for checking that radial average algorithms are working
    properly.

    The centre, intensity and radius position of the ring and disk can vary
    as a function of probe position, through the disk_x, disk_y, disk_r,
    disk_I, ring_x, ring_y, ring_r and ring_I arguments.
    In addition, the line width of the ring can be varied with ring_lw.

    There is also an elliptical ring, which can be added separately
    to the circular ring. This elliptical ring uses the ring_e_*
    arguments. It is disabled by default.

    The ring can be deactivated by setting ring_x=None.
    The disk can be deactivated by setting disk_x=None.
    The elliptical ring can be deactivated by setting ring_e_x=None.

    Parameters
    ----------
    probe_size_x, probe_size_y : int, default 10
        Size of the navigation dimension.
    image_size_x, image_size_y : int, default 50
        Size of the signal dimension.
    disk_x, disk_y : int or NumPy 2D-array, default 20
        Centre position of the disk. Either integer or NumPy 2-D array.
        See examples on how to make them the correct size.
        To deactivate the disk, set disk_x=None.
    disk_r : int or NumPy 2D-array, default 5
        Radius of the disk. Either integer or NumPy 2-D array.
        See examples on how to make it the correct size.
    disk_I : int or NumPy 2D-array, default 20
        Intensity of the disk, for each of the pixels.
        So if I=30, the each pixel in the disk will have a value of 30.
        Note, this value will change if blur=True or downscale=True.
    ring_x, ring_y : int or NumPy 2D-array, default 20
        Centre position of the ring. Either integer or NumPy 2-D array.
        See examples on how to make them the correct size.
        To deactivate the ring, set ring_x=None.
    ring_r : int or NumPy 2D-array, default 20
        Radius of the ring. Either integer or NumPy 2-D array.
        See examples on how to make it the correct size.
    ring_I : int or NumPy 2D-array, default 6
        Intensity of the ring, for each of the pixels.
        So if I=5, each pixel in the ring will have a value of 5.
        Note, this value will change if blur=True or downscale=True.
    ring_lw : int or NumPy 2D-array, default 0
        Line width of the ring. If ring_lw=1, the line will be 3 pixels wide.
        If ring_lw=2, the line will be 5 pixels wide.
    ring_e_x, ring_e_y : int or NumPy 2D-array, default 20
        Centre position of the elliptical ring. Either integer or
        NumPy 2-D array. See examples on how to make them the correct size.
        To deactivate the ring, set ring_x=None (which is the default).
    ring_e_semi_len0, ring_e_semi_len1 : int or NumPy 2D-array, default 15
        Semi lengths of the elliptical ring. Either integer or NumPy 2-D
        arrays. See examples on how to make it the correct size.
    ring_e_I : int or NumPy 2D-array, default 6
        Intensity of the elliptical ring, for each of the pixels.
        So if I=5, each pixel in the ring will have a value of 5.
        Note, this value will change if blur=True or downscale=True.
    ring_e_r : int or NumPy 2D-array, default 0
        Rotation of the elliptical ring, in radians.
    ring_e_lw : int or NumPy 2D-array, default 0
        Line width of the ring. If ring_lw=1, the line will be 3 pixels wide.
        If ring_lw=2, the line will be 5 pixels wide.
    blur : bool, default True
        If True, do a Gaussian blur of the disk.
    blur_sigma : int, default 1
        Sigma of the Gaussian blurring, if blur is True.
    downscale : bool, default True
        If True, use upscaling (then downscaling) to anti-alise the disk.
    add_noise : bool, default False
        Add Gaussian random noise.
    noise_amplitude : float, default 1
        The amplitude of the noise, if add_noise is True.
    lazy : bool, default False
        If True, the signal will be lazy
    lazy_chunks : tuple, optional
        Used if lazy is True, default (10, 10, 10, 10).

    Returns
    -------
    signal : HyperSpy Signal2D
        Signal with 2 navigation dimensions and 2 signal dimensions.

    Examples
    --------
    >>> from pyxem.data.dummy_data import make_diffraction_test_data as mdtd
    >>> s = mdtd.generate_4d_data(show_progressbar=False)
    >>> s.plot()

    Using more arguments

    >>> s = mdtd.generate_4d_data(probe_size_x=20, probe_size_y=30,
    ...         image_size_x=50, image_size_y=90,
    ...         disk_x=30, disk_y=70, disk_r=9, disk_I=30,
    ...         ring_x=35, ring_y=65, ring_r=20, ring_I=10,
    ...         blur=False, downscale=False, show_progressbar=False)

    Adding some Gaussian random noise

    >>> s = mdtd.generate_4d_data(add_noise=True, noise_amplitude=3,
    ...         show_progressbar=False)

    Different centre positions for each probe position.
    Note the size=(20, 10), and probe_x=10, probe_y=20: size=(y, x).

    >>> import numpy as np
    >>> disk_x = np.random.randint(5, 35, size=(20, 10))
    >>> disk_y = np.random.randint(5, 45, size=(20, 10))
    >>> disk_I = np.random.randint(50, 100, size=(20, 10))
    >>> ring_x = np.random.randint(5, 35, size=(20, 10))
    >>> ring_y = np.random.randint(5, 45, size=(20, 10))
    >>> ring_r = np.random.randint(10, 15, size=(20, 10))
    >>> ring_I = np.random.randint(1, 30, size=(20, 10))
    >>> ring_lw = np.random.randint(1, 5, size=(20, 10))
    >>> s = mdtd.generate_4d_data(probe_size_x=10, probe_size_y=20,
    ...         image_size_x=40, image_size_y=50, disk_x=disk_x, disk_y=disk_y,
    ...         disk_I=disk_I, ring_x=ring_x, ring_y=ring_y, ring_r=ring_r,
    ...         ring_I=ring_I, ring_lw=ring_lw, show_progressbar=False)

    Do not plot the disk

    >>> s = mdtd.generate_4d_data(disk_x=None, show_progressbar=False)

    Do not plot the ring

    >>> s = mdtd.generate_4d_data(ring_x=None, show_progressbar=False)

    Plot only an elliptical ring

    >>> from numpy.random import randint, random
    >>> s = mdtd.generate_4d_data(
    ...        probe_size_x=10, probe_size_y=10,
    ...        disk_x=None, ring_x=None,
    ...        ring_e_x=randint(20, 30, (10, 10)),
    ...        ring_e_y=randint(20, 30, (10, 10)),
    ...        ring_e_semi_len0=randint(10, 20, (10, 10)),
    ...        ring_e_semi_len1=randint(10, 20, (10, 10)),
    ...        ring_e_r=random((10, 10))*np.pi,
    ...        ring_e_lw=randint(1, 3, (10, 10)))

    """
    if disk_x is None:
        plot_disk = False
    else:
        plot_disk = True
        if not isiterable(disk_x):
            disk_x = np.ones((probe_size_y, probe_size_x)) * disk_x
    if not isiterable(disk_y):
        disk_y = np.ones((probe_size_y, probe_size_x)) * disk_y
    if not isiterable(disk_r):
        disk_r = np.ones((probe_size_y, probe_size_x)) * disk_r
    if not isiterable(disk_I):
        disk_I = np.ones((probe_size_y, probe_size_x)) * disk_I

    if ring_x is None:
        plot_ring = False
    else:
        plot_ring = True
        if not isiterable(ring_x):
            ring_x = np.ones((probe_size_y, probe_size_x)) * ring_x
    if not isiterable(ring_y):
        ring_y = np.ones((probe_size_y, probe_size_x)) * ring_y
    if not isiterable(ring_r):
        ring_r = np.ones((probe_size_y, probe_size_x)) * ring_r
    if not isiterable(ring_I):
        ring_I = np.ones((probe_size_y, probe_size_x)) * ring_I
    if not isiterable(ring_lw):
        ring_lw = np.ones((probe_size_y, probe_size_x)) * ring_lw

    if ring_e_x is None:
        plot_ring_e = False
    else:
        plot_ring_e = True
        if not isiterable(ring_e_x):
            ring_e_x = np.ones((probe_size_y, probe_size_x)) * ring_e_x
    if not isiterable(ring_e_y):
        ring_e_y = np.ones((probe_size_y, probe_size_x)) * ring_e_y
    if not isiterable(ring_e_semi_len0):
        ring_e_semi_len0 = np.ones((probe_size_y, probe_size_x)) * ring_e_semi_len0
    if not isiterable(ring_e_semi_len1):
        ring_e_semi_len1 = np.ones((probe_size_y, probe_size_x)) * ring_e_semi_len1
    if not isiterable(ring_e_I):
        ring_e_I = np.ones((probe_size_y, probe_size_x)) * ring_e_I
    if not isiterable(ring_e_lw):
        ring_e_lw = np.ones((probe_size_y, probe_size_x)) * ring_e_lw
    if not isiterable(ring_e_r):
        ring_e_r = np.ones((probe_size_y, probe_size_x)) * ring_e_r

    signal_shape = (probe_size_y, probe_size_x, image_size_y, image_size_x)
    s = Diffraction2D(np.zeros(shape=signal_shape))
    for i in tqdm(s, desc="Make test data", disable=not show_progressbar):
        index = s.axes_manager.indices[::-1]
        test_data = MakeTestData(
            size_x=image_size_x,
            size_y=image_size_y,
            default=False,
            blur=blur,
            blur_sigma=blur_sigma,
            downscale=downscale,
        )
        if plot_disk:
            dx, dy, dr = disk_x[index], disk_y[index], disk_r[index]
            dI = disk_I[index]
            test_data.add_disk(dx, dy, dr, intensity=dI)
        if plot_ring:
            rx, ry, rr = ring_x[index], ring_y[index], ring_r[index]
            rI, rLW = ring_I[index], ring_lw[index]
            test_data.add_ring(rx, ry, rr, intensity=rI, lw_pix=rLW)
        if plot_ring_e:
            rex, rey = ring_e_x[index], ring_e_y[index]
            resl0, resl1 = ring_e_semi_len0[index], ring_e_semi_len1[index]
            reI, reLW, rer = ring_e_I[index], ring_e_lw[index], ring_e_r[index]
            test_data.add_ring_ellipse(
                x0=rex,
                y0=rey,
                semi_len0=resl0,
                semi_len1=resl1,
                rotation=rer,
                intensity=reI,
                lw_r=reLW,
            )
        s.data[index][:] = test_data.signal.data[:]
        if add_noise:
            s.data[index][:] += (
                np.random.random(size=(image_size_y, image_size_x)) * noise_amplitude
            )
    s.axes_manager.indices = [0] * s.axes_manager.navigation_dimension
    if lazy:
        if lazy_chunks is None:
            lazy_chunks = 10, 10, 10, 10
        data_lazy = da.from_array(s.data, lazy_chunks)
        s = LazyDiffraction2D(data_lazy)
    return s


def _make_4d_peak_array_test_data(xf, yf, semi0, semi1, rot, nt=20):
    """Get a 4D NumPy array with peak_array test data.

    Parameters
    ----------
    xf, yf : scalar, 2D NumPy array
        Centre position of the ellipse. The size of the xf array gives the
        size of the peak_array.
    semi0, semi1 : scalar, 2D NumPy array
        Semi length of the ellipse, of rot is 0 semi0 is the x-direction
        semi length, and semi1 the y-direction.
    rot : scalar, 2D NumPy array
        Rotation in radians.
    nt : scalar
        Number of points in the ellipse.

    Returns
    -------
    peak_array : NumPy 4D array

    Examples
    --------
    >>> import pyxem as pxm
    >>> from pyxem.data.dummy_data import make_diffraction_test_data as mdtd
    >>> xf = np.random.randint(65, 70, size=(4, 5))
    >>> yf = np.random.randint(115, 120, size=(4, 5))
    >>> semi0 = np.random.randint(35, 40, size=(4, 5))
    >>> semi1 = np.random.randint(45, 50, size=(4, 5))
    >>> rot = np.random.random(size=(4, 5)) * 0.2
    >>> peak_array = mdtd._make_4d_peak_array_test_data(
    ...        xf, yf, semi0, semi1, rot)
    >>> s = pxm.signals.Diffraction2D(np.zeros(shape=(4, 5, 200, 210)))

    """
    peak_array = np.empty_like(xf, dtype=object)
    for iy, ix in np.ndindex(peak_array.shape):
        params = (xf[iy, ix], yf[iy, ix], semi0[iy, ix], semi1[iy, ix], rot[iy, ix], nt)
        ellipse_points = ret.make_ellipse_data_points(*params)
        peak_array[iy, ix] = np.fliplr(ellipse_points)
    return peak_array


class DiffractionTestImage:
    def __init__(
        self,
        disk_r=7,
        blur=2,
        image_x=256,
        image_y=256,
        rotation=0,
        diff_intensity_reduction=1.0,
        intensity_noise=0.5,
    ):
        """Make an artificial diffraction image, similar to NBED data.

        This class is for creating images which are similar to
        nanobeam electron diffraction patterns, with functionality for
        adding diffraction disks, Lorentzian background and intensity noise.

        Can be combined with the DiffractionTestDataset class, for creating
        4-dimensional datasets. Similar to the ones acquired with
        fast pixelated direct electron detectors in
        scanning transmission electron microscopy.

        Parameters
        ----------
        disk_r : int
            Radius for all of the disks added. Default 7
        blur : scalar
            Amount of Gaussian blur applied to the image.
            Use False to disable blurring. Default 2.
        image_x, image_y : int
            Dimensions for the image. Default 256 for both.
        rotation : int
            Rotation of the image in relation to the centre point.
            For an image with the size (256, 256), the rotation will
            be around (128, 128).
        diff_intensity_reduction : scalar
            In diffraction patterns, the intensity of the disk are
            reduced with increasing scattering angle. This parameter
            emulates this, with decreasing the intensity of the disks
            as a function of distance from the centre point. Default 1.0.
            Set False to disable.
        intensity_noise : scalar
            The width of the Gaussian intensity noise added to the image.
            Set False to disable. Default 0.5.


        Examples
        --------
        >>> from pyxem.data.dummy_data import make_diffraction_test_data as mdtd
        >>> di = mdtd.DiffractionTestImage()
        >>> di.add_disk(x=128, y=128, intensity=10.)
        >>> di.add_cubic_disks(vx=20, vy=20, intensity=2., n=5)
        >>> di.add_background_lorentz()
        >>> s = di.get_signal()
        >>> s.plot()

        Get a slightly rotated version of the diffraction image

        >>> di.rotation = 10
        >>> s = di.get_signal()
        >>> s.plot()

        """
        self.disk_r = disk_r
        self.blur = blur
        self.image_x = image_x
        self.image_y = image_y
        self.rotation = rotation
        self.diff_intensity_reduction = diff_intensity_reduction
        self.intensity_noise = intensity_noise
        self._background_lorentz_width = False
        self._background_lorentz_intensity = None
        self._x_list = []
        self._y_list = []
        self._intensity_list = []

    def __repr__(self):
        return "<%s, disks:%s, r:%s, rot:%s, im:(%s,%s)>" % (
            self.__class__.__name__,
            len(self._x_list),
            self.disk_r,
            self.rotation,
            self.image_x,
            self.image_y,
        )

    def __copy__(self):
        d = DiffractionTestImage(
            disk_r=self.disk_r,
            blur=self.blur,
            image_x=self.image_x,
            image_y=self.image_y,
            rotation=self.rotation,
            diff_intensity_reduction=self.diff_intensity_reduction,
            intensity_noise=self.intensity_noise,
        )
        d._background_lorentz_width = self._background_lorentz_width
        d._background_lorentz_intensity = self._background_lorentz_intensity
        d._x_list = self._x_list.copy()
        d._y_list = self._y_list.copy()
        d._intensity_list = self._intensity_list.copy()
        return d

    def copy(self):
        return self.__copy__()

    def add_disk(self, x, y, intensity=1):
        if not isinstance(x, int):
            raise ValueError("x needs to be integer, not {0}".format(x))
        if not isinstance(y, int):
            raise ValueError("y needs to be integer, not {0}".format(y))
        self._x_list.append(x)
        self._y_list.append(y)
        self._intensity_list.append(intensity)

    def add_background_lorentz(self, width=10, intensity=5):
        self._background_lorentz_width = width
        self._background_lorentz_intensity = intensity

    def _get_diff_intensity_reduction(self, dr, i):
        r_max = np.hypot(self.image_x / 2, self.image_y / 2)
        if dr == 0.0:
            dr = 0.00000000001
        x = np.pi * dr * 1.5 / (r_max * self.diff_intensity_reduction)
        i_new = np.sin(x) / x * i
        if i_new < 0:
            i_new = 0
        return i_new

    def add_cubic_disks(self, vx, vy, intensity=1, n=1):
        """Add disks in a cubic pattern around the centre point of the image.

        Parameters
        ----------
        vx, vy : int
        intensity : scalar
        n : int
            Number of orders of diffraction disks.
            If n=1, 8 disks, if n=2, 24 disks.
        """
        cx, cy = self.image_x / 2, self.image_y / 2
        for px in range(-n, n + 1):
            for py in range(-n, n + 1):
                if not (px == 0 and py == 0):
                    x = int(round(vx * px + cx))
                    y = int(round(vy * py + cy))
                    self.add_disk(x, y, intensity=intensity)

    def _get_background_lorentz(self):
        width = self._background_lorentz_width
        intensity = self._background_lorentz_intensity
        x = np.linspace(-width, width, self.image_y)
        y = np.linspace(-width, width, self.image_x)
        YY, XX = np.meshgrid(y, x)
        YY = YY.astype(np.float32)
        XX = XX.astype(np.float32)
        b = 1 / (np.pi * (1 + np.hypot(YY, XX) ** 2)) * intensity
        return b

    def get_diffraction_test_image(self, dtype=np.float32):
        image_x, image_y = self.image_x, self.image_y
        cx, cy = image_x / 2, image_y / 2
        image = np.zeros((image_y, image_x), dtype=np.float32)
        iterator = zip(self._x_list, self._y_list, self._intensity_list)
        for x, y, i in iterator:
            if self.diff_intensity_reduction is not False:
                dr = np.hypot(x - cx, y - cy)
                i = self._get_diff_intensity_reduction(dr, i)
            image[y, x] = i
        disk = morphology.disk(self.disk_r, dtype=dtype)
        image = convolve2d(image, disk, mode="same")
        if self.rotation != 0:
            image = rotate(image, self.rotation, reshape=False)
        if self.blur != 0:
            image = gaussian_filter(image, self.blur)
        if self._background_lorentz_width is not False:
            image += self._get_background_lorentz()
        if self.intensity_noise is not False:
            noise = np.random.random((image_y, image_x)) * self.intensity_noise
            image += noise
        return image

    def get_signal(self):
        s = Diffraction2D(self.get_diffraction_test_image())
        return s

    def plot(self):
        s = self.get_signal()
        s.plot()


class DiffractionTestDataset:
    def __init__(
        self,
        probe_x=10,
        probe_y=10,
        detector_x=256,
        detector_y=256,
        noise=0.5,
        dtype=np.float32,
    ):
        """Make a 4-dimensional dataset similar to NBED.

        This class is for creating datasets which are similar to
        nanobeam electron diffraction patterns. It is used in combination
        with one or several DiffractionTestImage objects.

        Parameters
        ----------
        probe_x, probe_y : int
        detector_x, detector_y : int
        noise : scalar

        Examples
        --------
        >>> from pyxem.data.dummy_data import make_diffraction_test_data as mdtd
        >>> di = mdtd.DiffractionTestImage(intensity_noise=False)
        >>> di.add_disk(x=128, y=128, intensity=10.)
        >>> di.add_cubic_disks(vx=20, vy=20, intensity=2., n=5)
        >>> di.add_background_lorentz()
        >>> di_rot = di.copy()
        >>> di_rot.rotation = 10
        >>> dtd = mdtd.DiffractionTestDataset(10, 10, 256, 256)
        >>> position_array = np.ones((10, 10), dtype=bool)
        >>> position_array[:5] = False
        >>> dtd.add_diffraction_image(di, position_array)
        >>> dtd.add_diffraction_image(di_rot, np.invert(position_array))
        >>> s = dtd.get_signal()

        """
        self.data = np.zeros((probe_x, probe_y, detector_x, detector_y), dtype=dtype)
        self.probe_x = probe_x
        self.probe_y = probe_y
        self.detector_x = detector_x
        self.detector_y = detector_y
        self.noise = noise

    def __repr__(self):
        return "<%s, (%s)>" % (self.__class__.__name__, self.data.shape)

    def add_diffraction_image(self, diffraction_test_image, position_array=None):
        """Add a diffraction image to all or a subset of the dataset.

        See the class docstring for example on how to use this.

        Parameters
        ----------
        diffraction_test_image : DiffractionTestData
        position_array : numpy.ndarray
            Boolean array, specifying which positions in the dataset
            the diffraction_test_image should be added to. Must have two
            dimensions, and the same shape as (probe_x, probe_y).

        """
        probe_x, probe_y = self.probe_x, self.probe_y
        detector_x, detector_y = self.detector_x, self.detector_y
        image = diffraction_test_image.get_diffraction_test_image()
        if position_array is None:
            position_array = np.ones((probe_x, probe_y), dtype=bool)
        for ix, iy in np.ndindex(probe_x, probe_y):
            if position_array[ix, iy]:
                self.data[ix, iy, :, :] = image
                if self.noise is not False:
                    image_noise = np.random.random((detector_x, detector_y))
                    self.data[ix, iy, :, :] += image_noise * self.noise

    def get_signal(self):
        s = Diffraction2D(self.data)
        return s

    def plot(self):
        s = self.get_signal()
        s.plot()



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

"""Utils for plotting 2D Diffraction Patterns."""

import math
import matplotlib.pyplot as plt
from matplotlib.collections import QuadMesh
import hyperspy.api as hs
import numpy as np
from pyxem.utils.polar_transform_utils import (
    get_template_cartesian_coordinates,
    get_template_polar_coordinates,
    image_to_polar,
)
from pyxem.utils.diffraction import find_beam_center_blur
import pyxem.utils._beam_shift_tools as bst
from mpl_toolkits.axes_grid1.anchored_artists import AnchoredSizeBar
from scipy.ndimage import gaussian_filter

__all__ = [
    "plot_template_over_pattern",
]


def plot_template_over_pattern(
    pattern,
    simulation,
    ax=None,
    in_plane_angle=0.0,
    max_r=None,
    find_direct_beam=True,
    direct_beam_position=None,
    mirrored_template=False,
    coordinate_system="cartesian",
    marker_color="red",
    marker_type="x",
    size_factor=1.0,
    **kwargs
):
    """A quick utility function to plot a simulated pattern over an experimental image

    Parameters
    ----------
    pattern : 2D np.ndarray
        The diffraction pattern
    simulation : :class:`diffsims.sims.diffraction_simulation.DiffractionSimulation`
        The simulated diffraction pattern. It must be calibrated.
    axis : matplotlib.AxesSubplot, optional
        An axis object on which to plot. If None is provided, one will be created.
    in_plane_angle : float, optional
        An in-plane rotation angle to apply to the template in degrees
    max_r : float, optional
        Maximum radius to consider in the polar transform in pixel coordinates.
        Will only influence the result if ``coordinate_system="polar"``.
    find_direct_beam: bool, optional
        Roughly find the optimal direct beam position if it is not centered.
    direct_beam_position: 2-tuple
        The (x, y) position of the direct beam in pixel coordinates. Takes
        precedence over `find_direct_beam`
    mirrored_template: bool, optional
        Whether to mirror the given template
    coordinate_system : str, optional
        Type of coordinate system to plot the image and template in. Either
        ``cartesian`` or ``polar``
    marker_color : str, optional
        Color of the spot markers
    marker_type : str, optional
        Type of marker used for the spots
    size_factor : float, optional
        Scaling factor for the spots. See notes on size.
    **kwargs :
        See :func:`matplotlib.pyplot.imshow`

    Returns
    -------
    ax : matplotlib.AxesSubplot
        The axes object
    im : matplotlib.image.AxesImage
        The representation of the image on the axes
    sp : matplotlib.collections.PathCollection
        The scatter plot representing the diffraction pattern

    Notes
    -----
    The spot marker sizes are scaled by the square root of their intensity
    """
    if ax is None:
        _, ax = plt.subplots()
    if coordinate_system == "polar":
        pattern = image_to_polar(
            pattern,
            max_r=max_r,
            find_direct_beam=find_direct_beam,
            direct_beam_position=direct_beam_position,
        )
        x, y, intensities = get_template_polar_coordinates(
            simulation,
            in_plane_angle=in_plane_angle,
            max_r=max_r,
            mirrored=mirrored_template,
        )
    elif coordinate_system == "cartesian":
        if direct_beam_position is not None:
            c_x, c_y = direct_beam_position
        elif find_direct_beam:
            c_x, c_y = find_beam_center_blur(pattern, 1)
        else:
            c_y, c_x = pattern.shape[0] / 2, pattern.shape[1] / 2
        x, y, intensities = get_template_cartesian_coordinates(
            simulation,
            center=(c_x, c_y),
            in_plane_angle=in_plane_angle,
            window_size=(pattern.shape[1], pattern.shape[0]),
            mirrored=mirrored_template,
        )
        y = pattern.shape[0] - y
    else:
        raise NotImplementedError(
            "Only polar and cartesian are accepted coordinate systems"
        )
    im = ax.imshow(pattern, **kwargs)
    sp = ax.scatter(
        x,
        y,
        s=size_factor * np.sqrt(intensities),
        marker=marker_type,
        color=marker_color,
    )
    return (ax, im, sp)


def plot_beam_shift_color(
    signal,
    phase_rotation=0,
    indicator_rotation=0,
    only_phase=False,
    autolim=True,
    autolim_sigma=4,
    magnitude_limits=None,
    scalebar_size=None,
    ax=None,
    ax_indicator=None,
):
    """Make a matplotlib figure showing beam shift.

    Parameters
    ----------
    phase_rotation : float, default 0
        Changes the phase of the plotted data.
        Useful for correcting scan rotation.
    indicator_rotation : float, default 0
        Changes the color wheel rotation.
    only_phase : bool, default False
        If False, will plot both the magnitude and phase.
        If True, will only plot the phase.
    autolim : bool, default True
    autolim_sigma : float, default 4
    magnitude_limits : tuple of floats, default None
        Manually sets the value limits for the color signal.
        For this, autolim needs to be False.
    scalebar_size : int, optional
    ax : Matplotlib subplot, optional
    ax_indicator : Matplotlib subplot, optional
        If None, generate a new subplot for the indicator.
        If False, do not include an indicator

    Examples
    --------
    >>> s = pxm.data.dummy_data.get_simple_beam_shift_signal()
    >>> fig = pxm.utils.plotting.plot_beam_shift_color(s)
    >>> fig.savefig("simple_beam_shift_test_signal.png")

    Only plotting the phase

    >>> fig = pxm.utils.plotting.plot_beam_shift_color(s, only_phase=True)
    >>> fig.savefig("simple_beam_shift_test_signal.png")

    Matplotlib subplot as input

    >>> import matplotlib.pyplot as plt
    >>> fig, ax = plt.subplots()
    >>> ax_indicator = fig.add_subplot(331)
    >>> fig = pxm.utils.plotting.plot_beam_shift_color(
    ...     s, scalebar_size=10, ax=ax, ax_indicator=ax_indicator)

    """
    indicator_rotation = indicator_rotation + 60
    if ax is None:
        set_fig = True
        fig, ax = plt.subplots(1, 1, figsize=(7, 7))
    else:
        fig = ax.figure
        set_fig = False
    if only_phase:
        s = signal.get_phase_signal(rotation=phase_rotation)
    else:
        s = signal.get_magnitude_phase_signal(
            rotation=phase_rotation,
            autolim=autolim,
            autolim_sigma=autolim_sigma,
            magnitude_limits=magnitude_limits,
        )
    s.change_dtype("uint16")
    s.change_dtype("float64")
    extent = s.axes_manager.navigation_extent
    extent = [extent[0], extent[1], extent[3], extent[2]]
    ax.imshow(s.data / 65536.0, extent=extent)
    if ax_indicator is not False:
        if ax_indicator is None:
            ax_indicator = fig.add_subplot(331)
        _make_color_wheel(
            ax_indicator,
            rotation=indicator_rotation + phase_rotation,
            only_phase=only_phase,
        )
    ax.set_axis_off()
    if scalebar_size is not None:
        scalebar_label = "{0} {1}".format(scalebar_size, s.axes_manager[0].units)
        sb = AnchoredSizeBar(ax.transData, scalebar_size, scalebar_label, loc=4)
        ax.add_artist(sb)
    if set_fig:
        fig.subplots_adjust(0, 0, 1, 1)
    return fig


def _make_color_wheel(ax, rotation=None, only_phase=False):
    x, y = np.mgrid[-2.0:2.0:500j, -2.0:2.0:500j]
    r = (x**2 + y**2) ** 0.5
    t = np.arctan2(x, y)
    del x, y
    if rotation is not None:
        t += math.radians(rotation)
        t = (t + np.pi) % (2 * np.pi) - np.pi

    r_masked = np.ma.masked_where((2.0 < r) | (r < 1.0), r)
    r_masked -= 1.0

    mask = r_masked.mask
    r_masked.data[r_masked.mask] = r_masked.mean()
    rgb_array = bst._get_rgb_phase_magnitude_array(
        t, r_masked.data, only_phase=only_phase
    )
    rgb_array = np.dstack((rgb_array, np.invert(mask)))

    ax.imshow(rgb_array, interpolation="quadric", origin="lower")
    ax.set_axis_off()


def make_color_wheel_marker(rotation=None, offsets=None, scale=0.2, only_phase=False):
    """Create a color wheel marker for use in hyperspy.

    Parameters
    ----------
    rotation : float, optional
        Rotation of the color wheel in degrees. Default is None, which means no rotation.
    offsets : list of lists, optional
        Offsets for the marker in axes coordinates. Default is [[0.8, 0.8]].
    scale : float, optional
        Scale factor for the marker size. Default is 0.2.
    """
    if offsets is None:
        offsets = [[0.8, 0.8]]
    x, y = np.mgrid[-2.0:2.0:360j, -2.0:2.0:360j]
    r = (x**2 + y**2) ** 0.5
    t = np.arctan2(x, y)
    if rotation is not None:
        t += math.radians(rotation)
        t = (t + np.pi) % (2 * np.pi) - np.pi

    r_masked = np.ma.masked_where((2.0 < r) | (r < 1.0), r)
    r_masked -= 1.0

    r_masked.data[r_masked.mask] = r_masked.mean()
    rgb_array = bst._get_rgb_phase_magnitude_array(
        t, r_masked.data, only_phase=only_phase
    )
    mask = (
        np.ma.masked_where((2.0 < r) | (r < 1.05), r) - 1.0
    ).mask  # slightly larger...
    alpha = np.invert(mask).astype(np.float32)
    alpha = gaussian_filter(alpha, sigma=1.5)  # Smooth the alpha channel
    rgb_array = np.dstack((rgb_array, alpha))  # Create RGBA array

    coords = np.stack([x, y], axis=-1)
    coords = coords * scale / 4  # Scale the coordinates
    quad = hs.plot.markers.Markers(
        collection=QuadMesh,
        coordinates=coords,
        array=rgb_array[1:, 1:],
        cmap="hsv",
        transform="axes",
        offset_transform="axes",
        offsets=offsets,
        alpha=1,
        antialiaseds=[
            True,
        ],
    )
    return quad



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

# For 0.20.0 we should update the code to call these functions from utils._dask, where they should be made private

"""Utils for Dask."""


from pyxem.utils._deprecated import deprecated
import scipy.ndimage as ndi


@deprecated(since="0.18.0", removal="0.20.0")
def align_single_frame(image, shifts, **kwargs):
    temp_image = ndi.shift(image, shifts[::-1], **kwargs)
    return temp_image


@deprecated(since="0.18.0", removal="0.20.0")
def get_signal_dimension_chunk_slice_list(chunks):
    """Convenience function for getting the signal chunks as slices

    The slices are assumed to be used on a HyperSpy signal object.
    Thus the input will be in the Dask chunk order (y, x), while the
    output will be in the HyperSpy order (x, y).

    """
    chunk_slice_raw_list = da.core.slices_from_chunks(chunks[-2:])
    chunk_slice_list = []
    for chunk_slice_raw in chunk_slice_raw_list:
        chunk_slice_list.append((chunk_slice_raw[1], chunk_slice_raw[0]))
    return chunk_slice_list


@deprecated(since="0.18.0", removal="0.20.0")
def get_signal_dimension_host_chunk_slice(x, y, chunks):
    chunk_slice_list = get_signal_dimension_chunk_slice_list(chunks)
    for chunk_slice in chunk_slice_list:
        x_slice, y_slice = chunk_slice
        if y_slice.start <= y < y_slice.stop:
            if x_slice.start <= x < x_slice.stop:
                return chunk_slice
    return False



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np
import hyperspy.api as hs

from hyperspy.signals import Signal1D
from pyxem.signals import InSituDiffraction2D


class TestTimeSeriesReconstruction:
    @pytest.fixture
    def insitu_data(self):
        dc = InSituDiffraction2D(data=np.ones(shape=(5, 2, 2, 25, 25)))
        dc.axes_manager.signal_axes[0].scale = 0.1
        dc.axes_manager.signal_axes[0].name = "kx"
        dc.axes_manager.signal_axes[1].scale = 0.1
        dc.axes_manager.signal_axes[1].name = "ky"
        return dc

    def test_roll_time_axis(self, insitu_data):
        rolled_data = insitu_data.roll_time_axis(0)
        assert (
            rolled_data.axes_manager.navigation_axes[2].size
            == insitu_data.axes_manager.navigation_axes[0].size
        )

    @pytest.mark.parametrize(
        "roi", [hs.roi.CircleROI(1, 1, 0.5), hs.roi.RectangularROI(0, 1, 1, 2), None]
    )
    def test_different_roi(self, insitu_data, roi):
        series = insitu_data.get_time_series(roi=roi)
        assert len(series.axes_manager.navigation_axes) == 1
        assert (
            series.axes_manager.navigation_axes[0].size
            == insitu_data.axes_manager.navigation_axes[2].size
        )
        assert (
            series.axes_manager.signal_axes[0].size
            == insitu_data.axes_manager.navigation_axes[0].size
        )
        assert (
            series.axes_manager.signal_axes[1].size
            == insitu_data.axes_manager.navigation_axes[1].size
        )

    def test_time_axis(self, insitu_data):
        series = insitu_data.get_time_series(
            roi=hs.roi.CircleROI(1, 1, 0.5), time_axis=0
        )
        assert (
            series.axes_manager.navigation_axes[0].size
            == insitu_data.axes_manager.navigation_axes[0].size
        )


class TestCorrelation:
    @pytest.fixture
    def insitu_data(self):
        dc = InSituDiffraction2D(data=np.random.rand(50, 10, 10, 4, 4))
        dc.axes_manager.signal_axes[0].scale = 0.1
        dc.axes_manager.signal_axes[0].name = "kx"
        dc.axes_manager.signal_axes[1].scale = 0.1
        dc.axes_manager.signal_axes[1].name = "ky"
        return dc

    def test_drift(self, insitu_data):
        shifts = insitu_data.get_drift_vectors()
        assert (
            shifts.axes_manager.navigation_axes[0].size
            == insitu_data.axes_manager.navigation_axes[2].size
        )
        assert shifts.axes_manager.signal_axes[0].size == 2

        assert isinstance(shifts, Signal1D)

    @pytest.mark.parametrize(
        "shifts",
        [
            Signal1D(np.zeros((50, 2))),
            Signal1D(
                np.repeat(np.linspace(0, 2, 50)[:, np.newaxis], repeats=2, axis=1)
            ),
            None,
        ],
    )
    def test_drift_corrected_g2_nonlazy(self, insitu_data, shifts):
        shifted_data = insitu_data.correct_real_space_drift(
            shifts=shifts, lazy_result=False
        )
        assert isinstance(shifted_data, InSituDiffraction2D)

        g2 = shifted_data.get_g2_2d_kresolved()
        assert g2.axes_manager.signal_axes[-1].size == 50
        mean_g2 = g2.isig[:, :, 1:].mean().data
        num_index = ~np.isreal(mean_g2)
        np.testing.assert_allclose(
            np.ones((49, 4, 4))[num_index], mean_g2[num_index], atol=0.1
        )

    def test_drift_corrected_g2_lazy(self, insitu_data):
        shifts = Signal1D(
            np.repeat(np.linspace(0, 2, 50)[:, np.newaxis], repeats=2, axis=1)
        )
        lazy_data = insitu_data.as_lazy()
        shifted_data = lazy_data.correct_real_space_drift(shifts=shifts)
        assert shifted_data._lazy
        assert isinstance(shifted_data, InSituDiffraction2D)
        g2_lazy = shifted_data.get_g2_2d_kresolved()
        assert g2_lazy.axes_manager.signal_axes[-1].size == 50
        g2_lazy.compute()
        mean_g2 = g2_lazy.isig[:, :, 1:].mean().data
        num_index = ~np.isreal(mean_g2)
        np.testing.assert_allclose(
            np.ones((49, 4, 4))[num_index], mean_g2[num_index], atol=0.1
        )

    @pytest.mark.parametrize(
        "shifts",
        [
            Signal1D(np.zeros((50, 2))),
            Signal1D(
                np.repeat(np.linspace(0, 2, 50)[:, np.newaxis], repeats=2, axis=1)
            ),
            None,
        ],
    )
    def test_fast_drift_corrected_g2_nonlazy(self, insitu_data, shifts):
        shifted_data = insitu_data.correct_real_space_drift_fast(shifts=shifts)
        assert isinstance(shifted_data, InSituDiffraction2D)

        g2 = shifted_data.get_g2_2d_kresolved()
        assert g2.axes_manager.signal_axes[-1].size == 50
        mean_g2 = g2.isig[:, :, 1:-1].mean(axis=[-1, -2, -3]).data
        num_index = ~np.isreal(mean_g2)
        np.testing.assert_allclose(
            np.ones((10, 10))[num_index], mean_g2[num_index], atol=0.1
        )

    def test_fast_drift_corrected_g2_lazy(self, insitu_data):
        shifts = Signal1D(
            np.repeat(np.linspace(0, 2, 50)[:, np.newaxis], repeats=2, axis=1)
        )
        lazy_data = insitu_data.as_lazy()
        lazy_data.rechunk((25, 5, 5, 4, 4))
        with pytest.raises(Exception) as exc_info:
            lazy_data.correct_real_space_drift_fast(shifts=shifts)

        assert exc_info.match(
            "Spatial axes are chunked. Please rechunk signal or use 'correct_real_space_drift' "
            "instead"
        )
        lazy_data.rechunk((1, 10, 10, 4, 4))
        shifted_data = lazy_data.correct_real_space_drift_fast(shifts=shifts)
        assert shifted_data._lazy
        assert isinstance(shifted_data, InSituDiffraction2D)

        g2_lazy = shifted_data.get_g2_2d_kresolved()
        assert g2_lazy.axes_manager.signal_axes[-1].size == 50
        g2_lazy.compute()
        mean_g2 = g2_lazy.isig[:, :, 1:-1].mean(axis=[-1, -2, -3]).data
        num_index = ~np.isreal(mean_g2)
        np.testing.assert_allclose(
            np.ones((10, 10))[num_index], mean_g2[num_index], atol=0.1
        )

    @pytest.mark.parametrize("normalization", ["self", "split"])
    def test_g2_normalization(self, insitu_data, normalization):
        g2 = insitu_data.get_g2_2d_kresolved(normalization=normalization)
        mean_g2 = g2.isig[:, :, 1:-1].mean(axis=[-1, -2, -3]).data
        num_index = ~np.isreal(mean_g2)
        np.testing.assert_allclose(
            np.ones((10, 10))[num_index], mean_g2[num_index], atol=0.1
        )

    @pytest.mark.parametrize("trs", [np.linspace(0, 10, 25), 10])
    @pytest.mark.parametrize("bins", [(2, 2, 5), (1, 4, 1)])
    def test_g2_bin_resample_time(self, insitu_data, trs, bins):
        g2 = insitu_data.get_g2_2d_kresolved(
            k1bin=bins[0], k2bin=bins[1], tbin=bins[2], resample_time=trs
        )
        assert g2.axes_manager.signal_axes[0].size == int(4 / bins[0])
        assert g2.axes_manager.signal_axes[1].size == int(4 / bins[1])
        mean_g2 = g2.isig[:, :, 1:-1].mean(axis=[-1, -2, -3]).data
        num_index = ~np.isreal(mean_g2)
        np.testing.assert_allclose(
            np.ones((10, 10))[num_index], mean_g2[num_index], atol=0.1
        )

    def test_unrolled_time_axes(self, insitu_data):
        rolled_data = insitu_data.roll_time_axis(0)
        shifted_data = rolled_data.correct_real_space_drift(
            shifts=Signal1D(np.zeros((50, 2))), time_axis=1
        )
        assert (
            shifted_data.axes_manager.navigation_axes[2].size
            == insitu_data.axes_manager.navigation_axes[2].size
        )
        shifted_data_fast = rolled_data.correct_real_space_drift_fast(
            shifts=Signal1D(np.zeros((50, 2))), time_axis=1
        )
        assert (
            shifted_data_fast.axes_manager.navigation_axes[2].size
            == insitu_data.axes_manager.navigation_axes[2].size
        )
        g2 = rolled_data.get_g2_2d_kresolved(time_axis=1)
        assert (
            g2.axes_manager.signal_axes[-1].size
            == insitu_data.axes_manager.navigation_axes[2].size
        )
